@@ -746,7 +746,7 @@ impl Rearrange {
 
         // Check that RearrangeSpec is valid for input, prevent panic on rearrange.info().shape
         out.spec
-            .conform_to_input_shape(&node.info().shape, false)
+            .conform_to_input_shape(&node.info().shape)
             .with_context(|| {
                 format!(
                     "{}\nname={:?}, child={:?}",
@@ -757,7 +757,7 @@ impl Rearrange {
             })?;
 
         // check that known sizes of each axis divide the operand shape
-        let input_sizes: Vec<u64> = out
+        let input_sizes: Vec<usize> = out
             .spec
             .input_ints
             .iter()
@@ -774,12 +774,13 @@ impl Rearrange {
                     .product()
             })
             .collect();
-        let any_sizes_dont_divide = node
-            .info()
-            .shape
-            .iter()
-            .enumerate()
-            .any(|(i, l)| *l as u64 % input_sizes[i] != 0);
+        let any_sizes_dont_divide = node.info().shape.iter().enumerate().any(|(i, l)| {
+            if input_sizes[i] == 0 {
+                *l == 0
+            } else {
+                *l % input_sizes[i] != 0
+            }
+        });
 
         if input_sizes.len() != node.info().rank() || any_sizes_dont_divide {
             return Err(ConstructError::RearrangeWrongInputShape {
@@ -824,7 +825,7 @@ circuit_node_extra_impl!(Rearrange, self_hash_default);
 
 impl CircuitNodeComputeInfoImpl for Rearrange {
     fn compute_shape(&self) -> Shape {
-        self.conform_to_input_shape_spec(false).shapes().unwrap().1
+        self.conform_to_input_shape_spec().shapes().unwrap().1
     }
 }
 
@@ -867,11 +868,11 @@ impl CircuitNode for Rearrange {
     }
 
     fn eval_tensors(&self, tensors: &[Tensor], _device_dtype: &TorchDeviceDtype) -> Result<Tensor> {
-        einops_repeat(
-            &tensors[0],
-            self.spec.to_einops_string(false),
-            self.spec.letter_sizes(),
-        )
+        // avoid using self.spec.apply so we get shape errors correctly
+        let (string, letter_sizes) = self
+            .conform_to_input_shape_spec()
+            .to_einops_string_and_letter_sizes();
+        einops_repeat(&tensors[0], string, letter_sizes)
     }
 
     fn intermediate_cost_bound(&self) -> usize {
@@ -923,11 +924,12 @@ impl Rearrange {
     }
 
     #[staticmethod]
-    pub fn from_einops_string(node: CircuitRc, string: &str, name: Option<String>) -> Result<Self> {
+    pub fn from_string(node: CircuitRc, string: &str, name: Option<String>) -> Result<Self> {
         Rearrange::try_new(
             node,
-            RearrangeSpec::from_einops_string(string)
-                .context("failed to construct spec for Rearrange::from_einops_string")?,
+            string
+                .parse()
+                .context("failed to construct spec for Rearrange::new_string")?,
             name,
         )
     }
@@ -945,19 +947,12 @@ impl Rearrange {
         )
     }
 
-    pub fn conform_to_input_shape_spec(&self, coerce: bool) -> RearrangeSpec {
-        self.spec
-            .conform_to_input_shape(&self.node.info().shape.clone(), coerce)
-            .unwrap()
+    pub fn conform_to_input_shape_spec(&self) -> RearrangeSpec {
+        self.spec.conform_to_input_shape(self.node.shape()).unwrap()
     }
 
-    pub fn conform_to_input_shape(&self, coerce: bool) -> Self {
-        Rearrange::try_new(
-            self.node.clone(),
-            self.conform_to_input_shape_spec(coerce),
-            None,
-        )
-        .unwrap()
+    pub fn conform_to_input_shape(&self) -> Self {
+        Rearrange::try_new(self.node.clone(), self.conform_to_input_shape_spec(), None).unwrap()
     }
 }
 
