@@ -5,7 +5,7 @@ use macro_rules_attribute::apply;
 use pyo3::{exceptions::PyValueError, prelude::*};
 use rr_util::{
     pycall, python_error_exception,
-    rearrange_spec::{OpShape, OpSize},
+    rearrange_spec::{ExpandToSpecOrShape, OpShape, OpSize},
     sv,
     symbolic_size::SymbolicSizeProduct,
     tensor_util::{Shape, TensorAxisIndex, TensorIndex},
@@ -56,7 +56,6 @@ pub fn expand_node(circuit: CircuitRc, inputs: &Vec<CircuitRc>) -> Result<Circui
 
     match &**circuit {
         Circuit::SetSymbolicShape(_) => Ok(inputs[0].clone()), /* just return circuit, symbolic shapes will be added as needed! (TODO: maybe should keep?) */
-        // TODO: symbolic shape case for below!!!
         Circuit::Symbol(_) | Circuit::Scalar(_) => Ok(circuit.clone()),
         Circuit::Rearrange(node) => {
             let input_shape_non_batch = inputs[0].info().shape[batch_ranks[0]..]
@@ -64,27 +63,38 @@ pub fn expand_node(circuit: CircuitRc, inputs: &Vec<CircuitRc>) -> Result<Circui
                 .cloned()
                 .collect();
 
-            let new_input = SetSymbolicShape::some_set_and_symbolic_neq(
-                inputs[0].clone(),
-                iter::repeat(OpSize::NONE)
-                    .take(batch_ranks[0])
-                    .chain(node.spec.forced_input_shape(true))
-                    .collect(),
-                None,
-            )
-            .with_context(|| {
-                format!(
-                    "failed to set symbolic in expand rearrange, old rearrange={:?}",
-                    node
-                )
-            })?;
-
-            let new_spec = node
+            // TODO: test me better!
+            let (new_input, expanded_spec) = match node
                 .spec
-                .conform_to_input_shape(&input_shape_non_batch, true)
-                .context("failed to conform rearrange to input shape in expand")?
-                .add_batch_dims(batch_ranks[0]);
-            Ok(Rearrange::nrc(new_input, new_spec, circuit.name_cloned()))
+                .expand_to_spec_or_shape(&input_shape_non_batch, false)
+                .expect("errors here have been handled above")
+            {
+                ExpandToSpecOrShape::Spec(spec) => (inputs[0].clone(), spec),
+                ExpandToSpecOrShape::SetShape(forced_shape) => {
+                    let new_input = SetSymbolicShape::some_set_and_symbolic_neq(
+                        inputs[0].clone(),
+                        iter::repeat(OpSize::NONE)
+                            .take(batch_ranks[0])
+                            .chain(forced_shape)
+                            .collect(),
+                        None,
+                    )
+                    .with_context(|| {
+                        format!(
+                            "failed to set symbolic in expand rearrange, old rearrange={:?}",
+                            node
+                        )
+                    })?;
+                    let expanded_spec = node
+                        .spec
+                        .expand_to(new_input.shape())
+                        .context("failed to expand rearrange to input shape in expand")?;
+                    (new_input, expanded_spec)
+                }
+            };
+
+            let new_spec = expanded_spec.add_batch_dims(batch_ranks[0]);
+            Rearrange::try_new(new_input, new_spec, circuit.name_cloned()).map(|x| x.rc())
         }
         // I think we just don't do any symbolic shape munging in Index/Scatter?
         Circuit::Index(node) => {
@@ -339,7 +349,7 @@ pub fn expand_node(circuit: CircuitRc, inputs: &Vec<CircuitRc>) -> Result<Circui
 
             // TODO: maybe dedup with parse and other places!
             let (arg_specs, nodes) = rest
-                .chunks(2)
+                .chunks_exact(2)
                 .zip(&node.spec.arg_specs)
                 .map(|(sym_inp, arg_spec)| {
                     if let [sym, inp] = sym_inp {
