@@ -23,6 +23,7 @@ use uuid::uuid;
 use crate::{
     circuit_node_auto_impl, circuit_node_extra_impl,
     circuit_node_private::{CircuitNodeComputeInfoImpl, CircuitNodeHashItems},
+    circuit_utils::OperatorPriority,
     new_rc_unwrap,
     prelude::*,
     Array, CachedCircuitInfo, HashBytes, PyCircuitBase,
@@ -666,19 +667,24 @@ impl CircuitNode for GeneralFunction {
 }
 
 impl CircuitNodeAutoName for GeneralFunction {
+    const PRIORITY: OperatorPriority = OperatorPriority::Function {};
+
     fn auto_name(&self, name: Option<String>) -> Option<String> {
+        // Never any parenthesis, so we don't even care checking if we need to add some
+        // but check out children_names_with_maybe_paren if you change the syntax of the general function
         name.or_else(|| {
             if self.children().any(|x| x.name().is_none()) {
                 None
             } else {
                 Some(
                     self.spec.name().clone()
-                        + " "
+                        + "("
                         + &self
                             .children()
                             .map(|x| Self::shorten_child_name(x.name().unwrap()))
                             .collect::<Vec<String>>()
-                            .join(", "),
+                            .join(", ")
+                        + ")",
                 )
             }
         })
@@ -1291,17 +1297,17 @@ fn test_index_general_function() -> Result<()> {
 #[base_exception(PyValueError)]
 #[derive(Error, Debug, Clone)]
 pub enum GeneralFunctionShapeError {
-    #[error("got={got} expected={expected}")]
+    #[error("got={got} expected={expected} ({e_name})")]
     WrongNumShapes { got: usize, expected: usize },
 
-    #[error("ndim={ndim} < num_non_batchable_output_dims={num_non_batchable_output_dims} + removed_from_end={removed_from_end}")]
+    #[error("ndim={ndim} < num_non_batchable_output_dims={num_non_batchable_output_dims} + removed_from_end={removed_from_end} ({e_name})")]
     NDimTooSmall {
         ndim: usize,
         num_non_batchable_output_dims: usize,
         removed_from_end: usize,
     },
 
-    #[error("x_shape={x_shape:?} index_shape={index_shape:?} batch_x={batch_x}")]
+    #[error("x_shape={x_shape:?} index_shape={index_shape:?} batch_x={batch_x} ({e_name})")]
     IndexShapeInvalid {
         x_shape: Shape,
         index_shape: Shape,
