@@ -47,7 +47,7 @@ pub use set_symbolic_size::SetSymbolicShape;
 use thiserror::Error;
 pub use variable_nodes::{DiscreteVar, StoredCumulantVar, Tag};
 
-use crate::{module::inline_all_modules, print::print_circuit_stats};
+use crate::{module::substitute_all_modules, print::print_circuit_stats};
 pub mod circuit_utils;
 pub mod computational_node;
 pub mod constant;
@@ -551,7 +551,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
     where
         Self: Clone,
     {
-        Rearrange::from_einops_string(self.crc(), string, name)
+        Rearrange::from_string(self.crc(), string, name)
     }
     fn add_suffix(self, suffix: Option<&str>) -> Self
     where
@@ -744,7 +744,7 @@ macro_rules! define_circuit_union_impl {
                         $(
                         stringify!($x) => Self::$x,
                         )*
-                        v => bail!(rr_util::tensor_util::ParseError::InvalidVariant {
+                        v => bail!($crate::parsing::ParseCircuitError::InvalidVariant {
                             v: v.to_owned()
                         }),
                     };
@@ -901,6 +901,8 @@ macro_rules! define_circuit {
     [$($x:ident),+ $(,)?] => {
         define_circuit_union_impl!(Circuit {$($x,)*});
 
+
+
         // you have to wrap pymethods in the paste for this to work due to how
         // pymethods proc macro works
         paste::paste! {
@@ -911,8 +913,15 @@ macro_rules! define_circuit {
                     arc_ref_clone(&self).into_op()
                 }
 
-                pub fn [<as_ $x:snake _unwrap>](&self) -> $crate::$x {
-                    arc_ref_clone(&self).into_unwrap()
+                pub fn [<as_ $x:snake _unwrap>](&self) -> PyResult<$x> {
+                    arc_ref_clone(&self).into_op().ok_or_else(|| {
+                        // could use fresh error type if we wanted
+                        PyErr::new::<pyo3::exceptions::PyTypeError, _>(format!(
+                            "expected {} but got {}",
+                            stringify!($x),
+                            self.type_tag()
+                        ))
+                    })
                 }
 
                 pub fn [<is_ $x:snake>](&self) -> bool {
@@ -1464,7 +1473,7 @@ pub fn evaluate_device_dtype(circ: CircuitRc, device_dtype: TorchDeviceDtypeOp)
             passed: device_dtype.clone(),
         })?
         .unwrap_or_defaults();
-    let circ = inline_all_modules(circ);
+    let circ = substitute_all_modules(circ);
     #[apply(cached_lambda)]
     #[key(circ.info().hash, HashBytes)]
     #[use_try]
