@@ -203,7 +203,7 @@ impl ModuleSpec {
 #[pyfunction]
 pub fn all_free(circuit: CircuitRc) -> Vec<Symbol> {
     let mut all_syms = Vec::new();
-    visit_circuit_unwrap(inline_all_modules(circuit), |x| {
+    visit_circuit_unwrap(substitute_all_modules(circuit), |x| {
         if let Some(sym) = x.as_symbol() {
             all_syms.push(sym.clone())
         }
@@ -667,7 +667,8 @@ impl Module {
     }
 
     #[staticmethod]
-    fn new_flat(nodes: Vec<CircuitRc>, spec: ModuleSpec, name: Option<String>) -> Result<Self> {
+    #[args(nodes = "*", name = "None")]
+    fn new_flat(spec: ModuleSpec, nodes: Vec<CircuitRc>, name: Option<String>) -> Result<Self> {
         Self::try_new(nodes, spec, name)
     }
 
@@ -731,7 +732,7 @@ impl Module {
 }
 
 #[pyfunction]
-pub fn inline_all_modules(circuit: CircuitRc) -> CircuitRc {
+pub fn substitute_all_modules(circuit: CircuitRc) -> CircuitRc {
     deep_map_unwrap(circuit, |c| match &**c {
         Circuit::Module(mn) => mn.substitute(None, None),
         _ => c.clone(),
