@@ -2,8 +2,12 @@ use std::{fmt::Debug, ops::Deref, sync::Arc};
 
 use anyhow::Result;
 use macro_rules_attribute::apply;
-use pyo3::{exceptions::PyTypeError, prelude::*, pyclass::CompareOp, PyObject};
-use rr_util::{py_types::use_rust_comp, python_error_exception};
+use pyo3::{exceptions::PyTypeError, prelude::*, pyclass::CompareOp, types::PyDict, PyObject};
+use rr_util::{
+    atr,
+    py_types::{use_rust_comp, SELF_MODULE},
+    pycall, python_error_exception,
+};
 use thiserror::Error;
 
 use crate::CircuitRc;
@@ -193,10 +197,19 @@ pub struct OpaqueIterateMatchResults {
     pub found: bool,
 }
 
-pub trait OpaqueIterativeMatcher: Debug {
+pub trait OpaqueIterativeMatcher: Debug + ToPyObject {
     fn opaque_match_iterate(&self, circuit: CircuitRc) -> Result<OpaqueIterateMatchResults>;
 }
 
+impl ToPyObject for OpaqueIterativeMatcherVal {
+    fn to_object(&self, py: Python<'_>) -> PyObject {
+        match self {
+            Self::Py(obj) => obj.clone(),
+            Self::Dyn(d) => d.to_object(py),
+        }
+    }
+}
+
 impl OpaqueIterativeMatcher for OpaqueIterativeMatcherVal {
     fn opaque_match_iterate(&self, circuit: CircuitRc) -> Result<OpaqueIterateMatchResults> {
         match self {
@@ -212,6 +225,26 @@ pub struct EndDepthOpaqueIterativeMatcher {
     pub depth: usize,
 }
 
+// why bother impl-ing if we also need to have python conversion?
+// We do this so that this so depth matcher can be used in rust without assuming python library is installed!
+// (same for below with Never matcher)
+impl ToPyObject for EndDepthOpaqueIterativeMatcher {
+    fn to_object(&self, _py: Python<'_>) -> PyObject {
+        assert_eq!(
+            self.depth, 0,
+            "shouldn't be called if depth has been updated"
+        );
+        Python::with_gil(|py| {
+            let dict = PyDict::new(py);
+            dict.set_item("end_depth", self.end_depth).unwrap();
+
+            OpaqueIterativeMatcherVal::noop_traversal_raw()
+                .call_method(py, "filter", (), Some(dict))
+                .unwrap()
+        })
+    }
+}
+
 impl OpaqueIterativeMatcher for EndDepthOpaqueIterativeMatcher {
     fn opaque_match_iterate(&self, _: CircuitRc) -> Result<OpaqueIterateMatchResults> {
         let found = self.depth < self.end_depth;
@@ -237,6 +270,12 @@ impl OpaqueIterativeMatcher for EndDepthOpaqueIterativeMatcher {
 #[derive(Clone, Debug)]
 pub struct NeverOpaqueIterativeMatcher;
 
+impl ToPyObject for NeverOpaqueIterativeMatcher {
+    fn to_object(&self, _: Python<'_>) -> PyObject {
+        pycall!(atr!(SELF_MODULE, IterativeMatcher, raw), (false,))
+    }
+}
+
 impl OpaqueIterativeMatcher for NeverOpaqueIterativeMatcher {
     fn opaque_match_iterate(&self, _: CircuitRc) -> Result<OpaqueIterateMatchResults> {
         Ok(OpaqueIterateMatchResults {
@@ -247,6 +286,10 @@ impl OpaqueIterativeMatcher for NeverOpaqueIterativeMatcher {
 }
 
 impl OpaqueIterativeMatcherVal {
+    pub fn noop_traversal_raw() -> PyObject {
+        atr!(SELF_MODULE, traversal, raw)
+    }
+
     pub fn for_end_depth(end_depth: usize) -> Self {
         Self::Dyn(Arc::new(EndDepthOpaqueIterativeMatcher {
             end_depth,
@@ -257,6 +300,14 @@ impl OpaqueIterativeMatcherVal {
     pub fn never() -> Self {
         Self::Dyn(Arc::new(NeverOpaqueIterativeMatcher))
     }
+
+    pub fn op_to_object(x: &Option<Self>) -> PyObject {
+        Python::with_gil(|py| {
+            x.as_ref()
+                .map(|x| x.to_object(py))
+                .unwrap_or_else(|| Self::noop_traversal_raw())
+        })
+    }
 }
 
 #[apply(python_error_exception)]
