@@ -8,8 +8,10 @@ use regex::Regex;
 use rr_util::{
     lru_cache::TensorCacheRrfs,
     python_error_exception,
-    symbolic_size::SymbolicSizeProduct,
-    tensor_util::{ParseError, Shape, TensorIndex, TorchDeviceDtypeOp},
+    symbolic_size::{SymbolicSizeProduct, SIZE_PROD_MATCH},
+    tensor_util::{
+        parse_numeric, ParseError, PyParseError, Shape, TensorIndex, TorchDeviceDtypeOp,
+    },
     util::{counts_g_1, flip_op_result, is_unique},
 };
 use rustc_hash::FxHashMap as HashMap;
@@ -33,6 +35,8 @@ pub struct Parser {
     #[pyo3(get, set)]
     pub tensors_as_random_device_dtype: TorchDeviceDtypeOp,
     #[pyo3(get, set)]
+    pub allow_hash_with_random: bool,
+    #[pyo3(get, set)]
     pub on_repeat_check_info_same: bool,
     #[pyo3(get, set)]
     pub module_check_all_args_present: bool,
@@ -46,6 +50,7 @@ impl Default for Parser {
             reference_circuits: Default::default(),
             tensors_as_random: false,
             tensors_as_random_device_dtype: TorchDeviceDtypeOp::NONE,
+            allow_hash_with_random: false,
             on_repeat_check_info_same: true,
             module_check_all_args_present: true,
             module_check_unique_arg_names: false,
@@ -60,6 +65,7 @@ impl Parser {
         reference_circuits_by_name = "vec![]",
         tensors_as_random = "Parser::default().tensors_as_random",
         tensors_as_random_device_dtype = "Parser::default().tensors_as_random_device_dtype",
+        allow_hash_with_random = "Parser::default().allow_hash_with_random",
         on_repeat_check_info_same = "Parser::default().on_repeat_check_info_same",
         module_check_all_args_present = "Parser::default().module_check_all_args_present",
         module_check_unique_arg_names = "Parser::default().module_check_unique_arg_names"
@@ -70,6 +76,7 @@ impl Parser {
         reference_circuits_by_name: Vec<CircuitRc>,
         tensors_as_random: bool,
         tensors_as_random_device_dtype: TorchDeviceDtypeOp,
+        allow_hash_with_random: bool,
         on_repeat_check_info_same: bool,
         module_check_all_args_present: bool,
         module_check_unique_arg_names: bool,
@@ -108,6 +115,7 @@ impl Parser {
             reference_circuits,
             tensors_as_random,
             tensors_as_random_device_dtype,
+            allow_hash_with_random,
             on_repeat_check_info_same,
             module_check_all_args_present,
             module_check_unique_arg_names,
@@ -142,11 +150,20 @@ impl Parser {
 }
 
 #[derive(Debug, Clone, Eq, Hash, PartialEq, PartialOrd, Ord)]
-enum CircuitIdent {
+pub enum CircuitIdent {
     Num(usize),
     Str(String),
 }
 
+impl IntoPy<PyObject> for CircuitIdent {
+    fn into_py(self, py: Python<'_>) -> PyObject {
+        match self {
+            Num(x) => x.into_py(py),
+            Str(x) => x.into_py(py),
+        }
+    }
+}
+
 use CircuitIdent::*;
 
 impl CircuitIdent {
@@ -176,7 +193,7 @@ impl Parser {
     ) -> Result<CircuitRc> {
         let circuits = self.parse_circuits(string, tensor_cache)?;
         if circuits.len() != 1 {
-            bail!(ParseError::ExpectedOneCircuitGotMultiple {
+            bail!(ParseCircuitError::ExpectedOneCircuitGotMultiple {
                 actual_num_circuits: circuits.len()
             })
         }
@@ -192,7 +209,8 @@ impl Parser {
 
         let tab_width: usize = 2;
         let mut partial_circuits: HashMap<CircuitIdent, PartialCirc> = HashMap::default();
-        let mut stack: Vec<CircuitIdent> = vec![];
+        // ident and if it's a new node
+        let mut stack: Vec<(CircuitIdent, bool)> = vec![];
         let mut top_level = vec![];
         const WHITESPACE: &str = r"[\s│└├‣]*";
         const NAME_INSIDE_MATCH: &str = r"(?:(?:\\')?[^']*)*";
@@ -210,10 +228,8 @@ impl Parser {
             let name_match_capture: String = format!("'({})'", NAME_INSIDE_MATCH);
             let name_match: String = format!("'{}'", NAME_INSIDE_MATCH);
             // first match named axis name, then size. Name is optional. We handle whitespace.
-            let maybe_symbolic_num = r"s?\d+";
-            let size_prod_match: String = format!(r"(?:{d}\s*\*\s*)*{d}", d = maybe_symbolic_num);
             let shape_axis_match: String =
-                format!(r"(?:{}\s*:\s*)?{}\s*", name_match, size_prod_match);
+                format!(r"(?:{}\s*:\s*)?{}\s*", name_match, *SIZE_PROD_MATCH);
             let trailing_comment_match = r"(?:#.*?)?"; // TODO: test this never matches stuff it's not supposed to...
             let full_regex = format!(
                 r"^({ws})(?:(\d+)|{nm})(?: {nm})?(?: \[((?:{sh},\s*)*(?:{sh})?)\])?(?: ([a-zA-Z]+))?(?: (.*?))?(?:{ws}){com}$",
@@ -238,12 +254,12 @@ impl Parser {
                 continue;
             }
 
-            let re_captures = re
-                .captures(line)
-                .ok_or_else(|| ParseError::RegexDidntMatch {
-                    line: line.to_owned(),
-                    regex_string: regex_string.clone(),
-                })?;
+            let re_captures =
+                re.captures(line)
+                    .ok_or_else(|| ParseCircuitError::RegexDidntMatch {
+                        line: line.to_owned(),
+                        regex_string: regex_string.clone(),
+                    })?;
             let num_spaces_base = re_captures
                 .get(prefix_whitespace_cap)
                 .expect("if regex matches, group should be present")
@@ -256,7 +272,7 @@ impl Parser {
             }
             let first_num_spaces = first_num_spaces.unwrap();
             if num_spaces_base < first_num_spaces {
-                bail!(ParseError::LessIndentationThanFirstItem {
+                bail!(ParseCircuitError::LessIndentationThanFirstItem {
                     first_num_spaces,
                     this_num_spaces_base: num_spaces_base,
                     first_line: first_line.unwrap(),
@@ -264,19 +280,22 @@ impl Parser {
                 });
             }
             let num_spaces = num_spaces_base - first_num_spaces;
+            let get_unindented_line = || line.chars().skip(first_num_spaces).collect();
             if num_spaces % tab_width != 0 {
-                bail!(ParseError::InvalidIndentation {
+                bail!(ParseCircuitError::InvalidIndentation {
                     tab_width,
                     spaces: num_spaces,
                     stack_indentation: stack.len(),
+                    line: get_unindented_line(),
                 });
             }
             let indentation_level = num_spaces / tab_width;
             if indentation_level > stack.len() {
-                bail!(ParseError::InvalidIndentation {
+                bail!(ParseCircuitError::InvalidIndentation {
                     tab_width,
                     spaces: num_spaces,
                     stack_indentation: stack.len(),
+                    line: get_unindented_line(),
                 });
             }
             stack.truncate(indentation_level);
@@ -289,7 +308,10 @@ impl Parser {
                 re_captures.get(name_ident_cap),
             ) {
                 (None, None) => unreachable!(),
-                (Some(c), None) => Num(c.as_str().parse().unwrap()),
+                (Some(c), None) => Num(c
+                    .as_str()
+                    .parse()
+                    .context("failed to parse serial number")?),
                 (None, Some(c)) => Str(unescape_name(c)),
                 (Some(_), Some(_)) => unreachable!(),
             };
@@ -306,10 +328,12 @@ impl Parser {
                     || re_captures.get(variant_cap).is_some()
                     || re_captures.get(extra_cap).is_some()
                 {
-                    bail!(ParseError::ReferenceCircuitNameFollowedByAdditionalInfo {
-                        reference_name: circuit_ident.as_str().unwrap().to_owned(),
-                        line: line.to_owned()
-                    });
+                    bail!(
+                        ParseCircuitError::ReferenceCircuitNameFollowedByAdditionalInfo {
+                            reference_name: circuit_ident.as_str().unwrap().to_owned(),
+                            line: get_unindented_line()
+                        }
+                    );
                 }
             } else {
                 let name = re_captures.get(name_cap).map(unescape_name).or_else(|| {
@@ -331,13 +355,10 @@ impl Parser {
                     }
 
                     let parse_axis_size = |x| {
-                        SymbolicSizeProduct::from_str(x)
-                            .context(concat!(
-                                "failed to parse number (including symbolic product) for shape",
-                                "\nthis is supposed to parse named axes but it currently doesn't!"
-                            ))?
-                            .try_into()
-                            .context("failed to convert size product to usize in parse of shape")
+                        SymbolicSizeProduct::parse_to_usize(x).context(concat!(
+                            "failed to parse number (including symbolic product) for shape",
+                            "\nthis is supposed to parse named axes but it currently doesn't!"
+                        ))
                     };
 
                     axis_strs
@@ -361,10 +382,13 @@ impl Parser {
                         PartialCirc {
                             name,
                             shape,
-                            variant: variant.ok_or_else(|| ParseError::RegexDidntMatchGroup {
-                                line: line.to_owned(),
-                                regex_string: regex_string.clone(),
-                                group: variant_cap,
+                            variant: variant.ok_or_else(|| {
+                                ParseCircuitError::RegexDidntMatchGroup {
+                                    group_name: "variant".to_owned(),
+                                    line: get_unindented_line(),
+                                    regex_string: regex_string.clone(),
+                                    group: variant_cap,
+                                }
                             })?,
                             extra: extra.unwrap_or_else(|| "".to_owned()),
                             children: vec![],
@@ -399,10 +423,10 @@ impl Parser {
                     fail_check!(variant);
                     fail_check!(extra);
                     if any_fail {
-                        bail!(ParseError::OnCircuitRepeatInfoIsNotSame {
+                        bail!(ParseCircuitError::OnCircuitRepeatInfoIsNotSame {
                             repeat_ident: format!("{:?}", circuit_ident),
                             err_strs,
-                            line: line.to_owned()
+                            line: get_unindented_line()
                         });
                     }
                 }
@@ -412,14 +436,21 @@ impl Parser {
             if stack.is_empty() {
                 top_level.push(circuit_ident.clone())
             }
-            if let Some(l) = stack.last() {
+            if let Some((l, is_new)) = stack.last() {
                 if let Str(s) = l {
                     if self.reference_circuits.contains_key(s) {
-                        bail!(ParseError::ReferenceCircuitHasChildrenInParseStr {
-                            reference_name: s.to_owned()
+                        bail!(ParseCircuitError::ReferenceCircuitHasChildren {
+                            reference_name: s.to_owned(),
+                            child_line: get_unindented_line(),
                         });
                     }
                 }
+                if !is_new {
+                    bail!(ParseCircuitError::RepeatedCircuitHasChildren {
+                        repeated_ident: l.clone(),
+                        child_line: get_unindented_line(),
+                    });
+                }
 
                 partial_circuits
                     .get_mut(l)
@@ -427,9 +458,7 @@ impl Parser {
                     .children
                     .push(circuit_ident.clone());
             }
-            if is_new_node {
-                stack.push(circuit_ident);
-            }
+            stack.push((circuit_ident, is_new_node));
         }
 
         let mut context = HashMap::default();
@@ -469,8 +498,8 @@ impl Parser {
             .deep_convert_partial_circ_children(ident, ps, children, tensor_cache)
             .with_context(|| {
                 format!(
-                    "in parse, failed to convert partial circuit with ident={:?}",
-                    ident
+                    "in parse, failed to convert ident={:?}, variant={}\nfull partial circuit: {:?}",
+                    ident, ps.variant, ps
                 )
             })?;
 
@@ -487,10 +516,10 @@ impl Parser {
     ) -> Result<CircuitRc> {
         let expected_k_children = |k| {
             if children.len() != k {
-                bail!(ParseError::WrongNumberChildren {
+                bail!(ParseCircuitError::WrongNumberChildren {
                     expected: k,
                     found: children.len(),
-                    ident: format!("{:?}", ident)
+                    ident: ident.clone()
                 })
             }
             Ok(())
@@ -509,76 +538,84 @@ impl Parser {
         let result = match ps.variant {
             T::Array => {
                 expected_k_children(0)?;
-                if self.tensors_as_random {
+                if self.tensors_as_random || ps.extra == "rand" {
+                    if ps.extra != "rand" && !self.allow_hash_with_random {
+                        // we allow "rand" even when self.tensors_as_random is true
+
+                        extra_should_be_empty().context(
+                            "self.tensors_as_random was passed, so array hashes should not be passed!",
+                        )?
+                    }
                     Array::randn_named(
-                        ps.shape.clone().ok_or(ParseError::ShapeNeeded {
-                            variant: ps.variant.to_string(),
+                        ps.shape.clone().ok_or(ParseCircuitError::ShapeNeeded {
+                            variant: ps.variant,
                         })?,
                         ps.name.clone(),
                         self.tensors_as_random_device_dtype.clone(),
                     )
                     .rc()
                 } else {
-                    Array::from_hash_prefix(ps.name.clone(), &ps.extra, tensor_cache)
-                        .context("parse array constant from hash prefix")?
-                        .rc()
+                    let out = Array::from_hash_prefix(ps.name.clone(), &ps.extra, tensor_cache)
+                        .context("failed to parse array constant from hash prefix")?
+                        .rc();
+                    if let Some(shape) = &ps.shape {
+                        if shape != out.shape() {
+                            bail!(ParseCircuitError::ArrayShapeLoadedFromHashDiffersFromProvidedShape {
+                                loaded_from_hash_shape : out.shape().clone(),
+                                provided_shape : shape.clone(),
+                            });
+                        }
+                    }
+                    out
                 }
             }
             T::Scalar => {
                 expected_k_children(0)?;
                 Scalar::nrc(
-                    ps.extra
-                        .parse::<f64>()
-                        .map_err(ParseError::from)
-                        .context("failed to parse out scalar number")?,
+                    parse_numeric(&ps.extra).context("failed to parse out number for Scalar")?,
                     ps.shape
                         .as_ref()
-                        .ok_or(ParseError::ShapeNeeded {
-                            variant: ps.variant.to_string(),
+                        .ok_or(ParseCircuitError::ShapeNeeded {
+                            variant: ps.variant,
                         })?
                         .clone(),
                     ps.name.clone(),
                 )
             }
             T::Add => {
-                if !ps.extra.is_empty() {
-                    bail!(ParseError::ExtraUnneededString {
-                        string: ps.extra.to_owned(),
-                    })
-                } else {
-                    Add::try_new(children, ps.name.clone())?.rc()
-                }
+                extra_should_be_empty()?;
+                Add::try_new(children, ps.name.clone())?.rc()
             }
             T::Concat => Concat::try_new(
                 children,
-                ps.extra
-                    .parse::<usize>()
-                    .map_err(ParseError::from)
-                    .context("failed to parse out concat axis")?,
+                parse_numeric(&ps.extra).context("failed to parse out concat axis")?,
                 ps.name.clone(),
             )?
             .rc(),
             T::Einsum => Einsum::from_einsum_string(&ps.extra, children, ps.name.clone())?.rc(),
             T::Rearrange => {
                 expected_k_children(1)?;
-                Rearrange::from_einops_string(children[0].clone(), &ps.extra, ps.name.clone())?.rc()
+                Rearrange::from_string(children[0].clone(), &ps.extra, ps.name.clone())?.rc()
             }
             T::Symbol => {
                 expected_k_children(0)?;
                 let shape = ps
                     .shape
                     .as_ref()
-                    .ok_or(ParseError::ShapeNeeded {
-                        variant: ps.variant.to_string(),
+                    .ok_or(ParseCircuitError::ShapeNeeded {
+                        variant: ps.variant,
                     })?
                     .clone();
                 if ps.extra.is_empty() {
                     Symbol::new_with_none_uuid(shape, ps.name.clone()).rc()
+                } else if ps.extra == "rand" {
+                    Symbol::new_with_random_uuid(shape, ps.name.clone()).rc()
                 } else {
                     Symbol::nrc(
                         shape,
-                        Uuid::from_str(&ps.extra).map_err(|_e| ParseError::Fail {
+                        Uuid::from_str(&ps.extra).map_err(|e| ParseError::InvalidUuid {
                             string: ps.extra.to_owned(),
+                            err_msg: e.to_string(),
                         })?,
                         ps.name.clone(),
                     )
@@ -603,8 +640,8 @@ impl Parser {
                     TensorIndex::from_bijection_string(&ps.extra, tensor_cache)?,
                     ps.shape
                         .as_ref()
-                        .ok_or(ParseError::ShapeNeeded {
-                            variant: ps.variant.to_string(),
+                        .ok_or(ParseCircuitError::ShapeNeeded {
+                            variant: ps.variant,
                         })?
                         .clone(),
                     ps.name.clone(),
@@ -614,7 +651,7 @@ impl Parser {
             T::Module => {
                 let arg_info = if ps.extra == "all_t" {
                     if children.len() % 2 != 1 {
-                        bail!(ParseError::ModuleInconsistentNumChildren {
+                        bail!(ParseCircuitError::ModuleInconsistentNumChildren {
                             name: ps.name.clone(),
                             extra: ps.extra.clone(),
                             num_children: children.len(),
@@ -630,7 +667,7 @@ impl Parser {
                         .map(|x| {
                             let num_chars = x.chars().count();
                             if num_chars != 2 {
-                                bail!(ParseError::Expected2CharsPerForModuleExtra {
+                                bail!(ParseCircuitError::Expected2CharsPerForModuleExtra {
                                     name: ps.name.clone(),
                                     num_chars,
                                     extra: ps.extra.clone(),
@@ -648,7 +685,7 @@ impl Parser {
                 };
 
                 if children.len() != 2 * arg_info.len() + 1 {
-                    bail!(ParseError::ModuleInconsistentNumChildren {
+                    bail!(ParseCircuitError::ModuleInconsistentNumChildren {
                         name: ps.name.clone(),
                         extra: ps.extra.clone(),
                         num_children: children.len(),
@@ -660,7 +697,7 @@ impl Parser {
                 let spec_circuit = children.pop().unwrap();
 
                 let (arg_specs, nodes) = rest
-                    .chunks(2)
+                    .chunks_exact(2)
                     .zip(arg_info)
                     .map(|(sym_inp, [batchable, expandable])| {
                         let (sym, inp) = if let [sym, inp] = sym_inp {
@@ -722,35 +759,37 @@ impl Parser {
                     .rc()
             }
             T::StoredCumulantVar => {
-                if let [cum_nums, uuid] = ps.extra.split("|").collect::<Vec<_>>()[..] {
-                    let keys = cum_nums
-                        .split(",")
-                        .map(|s| s.trim().parse::<usize>().map_err(ParseError::from))
-                        .collect::<Result<Vec<_>, _>>()
-                        .context("failed to parse keys for stored cumulant var")?;
-
-                    let uuid = Uuid::from_str(uuid).map_err(|e| ParseError::InvalidUuid {
-                        string: ps.extra.clone(),
-                        err_msg: e.to_string(),
-                    })?;
-                    if keys.len() != children.len() {
-                        bail!(ParseError::WrongNumberChildren {
-                            expected: keys.len(),
-                            found: children.len(),
-                            ident: format!("{:?}", ident)
-                        })
+                let (cum_nums, uuid) =
+                    if let [cum_nums, uuid] = ps.extra.split("|").collect::<Vec<_>>()[..] {
+                        (cum_nums, uuid)
                     } else {
-                        StoredCumulantVar::try_new(
-                            zip(keys, children).collect(),
-                            uuid,
-                            ps.name.clone(),
-                        )?
-                        .rc()
-                    }
-                } else {
-                    bail!(ParseError::Fail {
-                        string: ps.extra.clone(),
+                        bail!(ParseCircuitError::StoredCumulantVarExtraInvalid {
+                            extra: ps.extra.clone(),
+                        })
+                    };
+                let keys = cum_nums
+                    .split(",")
+                    .map(|s| parse_numeric(s.trim()))
+                    .collect::<Result<Vec<_>, _>>()
+                    .context("failed to parse keys for stored cumulant var")?;
+
+                let uuid = Uuid::from_str(uuid).map_err(|e| ParseError::InvalidUuid {
+                    string: ps.extra.clone(),
+                    err_msg: e.to_string(),
+                })?;
+                if keys.len() != children.len() {
+                    bail!(ParseCircuitError::WrongNumberChildren {
+                        expected: keys.len(),
+                        found: children.len(),
+                        ident: ident.clone()
                     })
+                } else {
+                    StoredCumulantVar::try_new(
+                        zip(keys, children).collect(),
+                        uuid,
+                        ps.name.clone(),
+                    )?
+                    .rc()
                 }
             }
             T::SetSymbolicShape => {
@@ -760,8 +799,8 @@ impl Parser {
                     children[0].clone(),
                     ps.shape
                         .as_ref()
-                        .ok_or(ParseError::ShapeNeeded {
-                            variant: ps.variant.to_string(),
+                        .ok_or(ParseCircuitError::ShapeNeeded {
+                            variant: ps.variant,
                         })?
                         .clone(),
                     ps.name.clone(),
@@ -787,3 +826,126 @@ pub enum ParseArgError {
     #[error("dup_idents={dup_idents:?} ({e_name})")]
     ReferenceCircuitDuplicateIdentifier { dup_idents: HashMap<String, usize> },
 }
+
+const REGEX_101_MSG: &str = "\nTry using regex101.com : )\n(TODO: auto generate regex test page via https://github.com/firasdib/Regex101/wiki/API#python-3)";
+
+#[apply(python_error_exception)]
+#[base_error_name(ParseCircuit)]
+#[base_exception(PyParseError)]
+#[derive(Error, Debug, Clone)]
+pub enum ParseCircuitError {
+    #[error(
+        "regex didn't match '{line}'\nregex='{regex_string}'{}\n({e_name})",
+        REGEX_101_MSG
+    )]
+    RegexDidntMatch { line: String, regex_string: String },
+
+    #[error(
+        "regex didn't match '{group_name}' group={group} for '{line}'\n{}\nregex='{regex_string}'{}\n({e_name})",
+        "Note: this error can also be caused by failing to lookup a reference circuit",
+        REGEX_101_MSG
+    )]
+    RegexDidntMatchGroup {
+        group_name: String,
+        line: String,
+        regex_string: String,
+        group: usize,
+    },
+
+    #[error("effectively negative indent!\n{}\n{}\n{}\n({e_name})",
+        format!("this_num_spaces_base={} < first_num_spaces={}", this_num_spaces_base, first_num_spaces),
+        format!("first_line='{}'", first_line),
+        format!("this_line='{}'", this_line)
+        )]
+    LessIndentationThanFirstItem {
+        this_num_spaces_base: usize,
+        first_num_spaces: usize,
+        first_line: String,
+        this_line: String,
+    },
+
+    #[error("Parsing invalid indentation, tab width {tab_width} num spaces {spaces} stack indentation {stack_indentation}\nline='{line}'\n({e_name})")]
+    InvalidIndentation {
+        spaces: usize,
+        tab_width: usize,
+        stack_indentation: usize,
+        line: String,
+    },
+
+    #[error("extra='{extra}' for variant={variant}, but expected empty ({e_name})")]
+    ExpectedNoExtraInfo { extra: String, variant: CircuitType },
+
+    #[error("Parsing wrong number of children, expected {expected} found {found}, ident={ident:?} ({e_name})")]
+    WrongNumberChildren {
+        expected: usize,
+        found: usize,
+        ident: CircuitIdent,
+    },
+
+    #[error("Parsing invalid circuit variant {v} ({e_name})")]
+    InvalidVariant { v: String },
+
+    #[error("Parsing shape needed but not provided on {variant} ({e_name})")]
+    ShapeNeeded { variant: CircuitType },
+
+    #[error("actual_num_circuits={actual_num_circuits} ({e_name})")]
+    ExpectedOneCircuitGotMultiple { actual_num_circuits: usize },
+
+    #[error("expected 't' or 'f' got='{got}' ({e_name})")]
+    InvalidTerseBool { got: String },
+
+    #[error("reference_name={reference_name} child_line='{child_line}' ({e_name})")]
+    ReferenceCircuitHasChildren {
+        reference_name: String,
+        child_line: String,
+    },
+
+    #[error("repeated_ident={repeated_ident:?} child_line='{child_line}' ({e_name})")]
+    RepeatedCircuitHasChildren {
+        repeated_ident: CircuitIdent,
+        child_line: String,
+    },
+
+    #[error("reference_name={reference_name} ({e_name})")]
+    ReferenceCircuitNameFollowedByAdditionalInfo {
+        reference_name: String,
+        line: String,
+    },
+
+    #[error(
+        "loaded_from_hash_shape={loaded_from_hash_shape:?} != provided_shape={provided_shape:?}.{}\n({e_name})",
+        " Note that you don't have to provide the shape if you provide a hash"
+    )]
+    ArrayShapeLoadedFromHashDiffersFromProvidedShape {
+        loaded_from_hash_shape: Shape,
+        provided_shape: Shape,
+    },
+
+    #[error("repeat_ident={repeat_ident} \n{err_strs}line='{line}' ({e_name})")]
+    OnCircuitRepeatInfoIsNotSame {
+        repeat_ident: String,
+        line: String,
+        err_strs: String,
+    },
+
+    #[error("num_chars={num_chars} extra='{extra}' name={name:?}\nmodule extra should look like 'tt,tf,ft' and we check for 2 chars per comma delimited item.\n({e_name})")]
+    Expected2CharsPerForModuleExtra {
+        name: Option<String>,
+        num_chars: usize,
+        extra: String,
+    },
+
+    #[error("num_children={num_children} != 2 * num_args={num_args:?} + 1\nname={name:?} extra={extra}\n{}\n{}\n{}\n({e_name})",
+    "modules have 1 child for the spec and then a symbol and a input circuit for each arg (2 * num_args + 1)",
+    "this implies the number of children is always odd",
+    "num_args is based on parsing the batchable and expandable info from 'extra' (see above) (in a form like 'tt,tf,ft' (Some(3)) or 'all_t' (None))")]
+    ModuleInconsistentNumChildren {
+        name: Option<String>,
+        num_children: usize,
+        extra: String,
+        num_args: Option<usize>,
+    },
+
+    #[error("extra='{extra}' invalid. Should contain | (TODO: example). ({e_name})")]
+    StoredCumulantVarExtraInvalid { extra: String },
+}
