@@ -11,7 +11,6 @@ use num_bigint::BigUint;
 use pyo3::{once_cell::GILLazy, prelude::*};
 use rr_util::{
     fn_struct, py_types::MaybeNotSet, python_println, symbolic_size::SymbolicSizeProduct,
-    tensor_util::ParseError,
 };
 use rustc_hash::FxHashMap as HashMap;
 
@@ -19,12 +18,13 @@ use crate::{
     circuit_utils::{count_nodes, total_flops},
     generalfunction::SpecTrait,
     opaque_iterative_matcher::{OpaqueIterativeMatcher, OpaqueIterativeMatcherVal},
+    parsing::ParseCircuitError,
     prelude::*,
     HashBytes,
 };
 
-// TODO: fix color scheme
-const COLOR_CODES: [usize; 15] = [30, 31, 32, 33, 34, 35, 36, 90, 91, 92, 93, 94, 95, 96, 97];
+// TODO: improve color scheme
+const COLOR_CODES: [usize; 14] = [31, 32, 33, 34, 35, 36, 90, 91, 92, 93, 94, 95, 96, 97];
 
 const BAR: &'static str = "│";
 const TEE: &'static str = "├";
@@ -50,6 +50,7 @@ pub struct PrintOptions {
     pub print_size_threshold: Option<usize>,
     #[pyo3(get, set)]
     pub force_use_serial_numbers: bool,
+    #[pyo3(get, set)]
     pub colorer: Option<CircuitToColorCode>,
 }
 
@@ -106,7 +107,7 @@ impl std::str::FromStr for TerseBool {
         let out = match s {
             "t" => true,
             "f" => false,
-            _ => bail!(ParseError::InvalidTerseBool { got: s.to_owned() }),
+            _ => bail!(ParseCircuitError::InvalidTerseBool { got: s.to_owned() }),
         };
         Ok(TerseBool(out))
     }
@@ -117,7 +118,7 @@ impl TryFrom<char> for TerseBool {
         let out = match value {
             't' => true,
             'f' => false,
-            _ => bail!(ParseError::InvalidTerseBool {
+            _ => bail!(ParseCircuitError::InvalidTerseBool {
                 got: value.to_string()
             }),
         };
@@ -227,6 +228,16 @@ impl PrintOptions {
         )
     }
 
+    #[getter(traversal)]
+    pub fn py_get_traversal(&self) -> PyObject {
+        OpaqueIterativeMatcherVal::op_to_object(&self.traversal)
+    }
+
+    #[setter(traversal)]
+    pub fn py_set_traversal(&mut self, traversal: Option<OpaqueIterativeMatcherVal>) {
+        self.traversal = traversal
+    }
+
     #[staticmethod]
     pub fn compiler_default() -> PrintOptions {
         Self {
@@ -313,7 +324,7 @@ impl PrintOptions {
                 Circuit::Scalar(scalar) => {
                     format!("{:.}", scalar.value)
                 }
-                Circuit::Rearrange(rearrange) => rearrange.spec.to_einops_string(true),
+                Circuit::Rearrange(rearrange) => rearrange.spec.to_string(),
                 Circuit::Einsum(einsum) => einsum.get_spec().to_einsum_string(),
                 Circuit::Index(index) => {
                     if self.bijection {
@@ -398,7 +409,7 @@ impl PrintOptions {
         });
         if let Some(threshold) = self.print_size_threshold {
             if circ.info().numel() > BigUint::from(threshold) && !circ.is_array() {
-                result.push_str(&format!(" # {}", color(&oom_fmt(circ.info().numel()), 1)));
+                result.push_str(&format!(" # {}", color(&oom_fmt(circ.info().numel()), 0)));
             }
         }
 
@@ -675,15 +686,15 @@ fn init_print_options() -> PrintOptions {
 pub fn circuit_to_size_color_code(circuit: CircuitRc) -> usize {
     let size = circuit.info().numel_usize();
     if size == usize::MAX {
-        3
-    } else if size > 10_000_000_000 {
-        4
-    } else if size > 300_000_000 {
-        1
-    } else if size > 300_000 {
         2
-    } else {
+    } else if size > 10_000_000_000 {
+        3
+    } else if size > 300_000_000 {
         0
+    } else if size > 300_000 {
+        1
+    } else {
+        4
     }
 }
 
