@@ -4,7 +4,7 @@ use std::{
     hash::{Hash, Hasher},
 };
 
-use anyhow::Result;
+use anyhow::{anyhow, bail, Context, Result};
 use pyo3::prelude::*;
 use rand::Rng;
 use rr_util::{fn_struct, py_types::MaybeNotSet, simple_default, util::hash_to_string};
@@ -223,13 +223,13 @@ impl PrintHtmlOptions {
     #[new]
     #[args(
         shape_only_when_necessary = "PrintHtmlOptions::default().shape_only_when_necessary",
-        traversal = "None",
+        traversal = "PrintHtmlOptions::default().traversal.unwrap()",
         display_copy_button = "PrintHtmlOptions::default().display_copy_button",
-        colorer = "None",
+        colorer = "PrintHtmlOptions::default().colorer.unwrap()",
         primary_color = "PrintHtmlOptions::default().primary_color",
         display_serial_numbers = "PrintHtmlOptions::default().display_serial_numbers",
         keep_all_cells = "PrintHtmlOptions::default().keep_all_cells",
-        overall_traversal = "None"
+        overall_traversal = "PrintHtmlOptions::default().overall_traversal.unwrap()"
     )] // None are mapped to the default
     pub fn py_new(
         shape_only_when_necessary: bool,
@@ -243,15 +243,13 @@ impl PrintHtmlOptions {
     ) -> Result<Self> {
         let result = Self {
             shape_only_when_necessary,
-            traversal: Some(traversal.unwrap_or(PrintHtmlOptions::default().traversal.unwrap())),
+            traversal,
             display_copy_button,
-            colorer: Some(colorer.unwrap_or(PrintHtmlOptions::default().colorer.unwrap())),
+            colorer,
             primary_color,
             display_serial_numbers,
             keep_all_cells,
-            overall_traversal: Some(
-                overall_traversal.unwrap_or(PrintHtmlOptions::default().overall_traversal.unwrap()),
-            ),
+            overall_traversal,
         };
         Ok(result)
     }
@@ -294,6 +292,29 @@ impl PrintHtmlOptions {
         }
     }
 
+    #[getter(traversal)]
+    pub fn py_get_traversal(&self) -> PyObject {
+        OpaqueIterativeMatcherVal::op_to_object(&self.traversal)
+    }
+
+    #[setter(traversal)]
+    pub fn py_set_traversal(&mut self, traversal: Option<OpaqueIterativeMatcherVal>) {
+        self.traversal = traversal
+    }
+
+    #[getter(overall_traversal)]
+    pub fn py_get_overall_traversal(&self) -> PyObject {
+        OpaqueIterativeMatcherVal::op_to_object(&self.overall_traversal)
+    }
+
+    #[setter(overall_traversal)]
+    pub fn py_set_overall_traversal(
+        &mut self,
+        overall_traversal: Option<OpaqueIterativeMatcherVal>,
+    ) {
+        self.overall_traversal = overall_traversal
+    }
+
     #[args(circuits = "*")]
     pub fn print(&self, circuits: Vec<CircuitRc>) -> Result<()> {
         let html = self.repr(circuits)?;
@@ -345,8 +366,8 @@ impl PrintHtmlOptions {
         html.push_str(
             &data
                 .iter()
-                .map(|(h, json_data)| hash_to_string(h) + ":" + &json_data.to_string())
-                .collect::<Vec<String>>()
+                .map(|(h, json_data)| Ok(hash_to_string(h) + ":" + &json_data.to_string()?))
+                .collect::<Result<Vec<String>>>()?
                 .join(","),
         );
         html.push_str("};");
@@ -378,7 +399,7 @@ impl PrintHtmlOptions {
 
         html.push_str(&format!(
             "--primary: {};",
-            make_str_safe(&self.primary_color)
+            make_str_css_safe(&self.primary_color)?
         ));
         html.push_str(CSS);
 
@@ -387,39 +408,12 @@ impl PrintHtmlOptions {
         ));
         html.push_str(JS);
 
-        // If keep_all_cells, variables in all cells need to have different names.
-        // To do so, the seed is appended to variable names.
-        // Note: conflict with circuit name could be problematic since we do a bulk replacement
-        // which is why variable are named with ___js___
         if self.keep_all_cells {
-            html = html
-                .replace(
-                    "expandClonestoggle___js___",
-                    &format!("expandClonestoggle___js___{seed}"),
-                )
-                .replace(
-                    "toExpandtoggle___js___",
-                    &format!("toExpandtoggle___js___{seed}"),
-                )
-                .replace("datatoggle___js___", &format!("datatoggle___js___{seed}"))
-                .replace("rootstoggle___js___", &format!("rootstoggle___js___{seed}"))
-                .replace(
-                    "displayCopyButtontoggle___js___",
-                    &format!("displayCopyButtontoggle___js___{seed}"),
-                )
-                .replace("seedtoggle___js___", &format!("seedtoggle___js___{seed}"))
-                .replace(
-                    "addChildtoggle___js___",
-                    &format!("addChildtoggle___js___{seed}"),
-                )
-                .replace(
-                    "toggletoggle___js___",
-                    &format!("toggletoggle___js___{seed}"),
-                )
-                .replace(
-                    "toggleSiblingsBoldtoggle___js___",
-                    &format!("toggleSiblingsBoldtoggle___js___{seed}"),
-                );
+            // If keep_all_cells, variables in all cells need to have different names.
+            // To do so, the seed needs to be in variable names which shouldn't be overwritten.
+            // These variables have ___js___ as suffix.
+            // Note: conflict with circuit name could be problematic.
+            html = html.replace("___js___", &format!("{seed}"));
         }
 
         Ok(html)
@@ -518,21 +512,29 @@ pub struct NodeJsonData {
     children: Vec<String>,
 }
 
-fn make_str_safe(s: &String) -> String {
-    s.replace(";", "")
-        .replace("\"", "")
-        .replace("<", "&lt")
-        .replace(">", "&gt")
+fn make_str_safe(s: &String) -> Result<String> {
+    if s.contains("___js___") {
+        bail!(anyhow!("___js___ is a reserved name. Don't use it."))
+    }
+    if s.contains("\"") {
+        bail!(anyhow!("Can't use quotes in string printed to html."))
+    }
+    Ok(s.replace("<", "&lt").replace(">", "&gt"))
+}
+fn make_str_css_safe(s: &String) -> Result<String> {
+    return Ok(make_str_safe(s)
+        .context("wrong color format (no quotes allowed)")?
+        .replace(";", ""));
 }
 
-impl ToString for NodeJsonData {
-    fn to_string(&self) -> String {
+impl NodeJsonData {
+    fn to_string(&self) -> Result<String> {
         let mut r = String::from("{name:\"");
-        r.push_str(&make_str_safe(&self.name));
+        r.push_str(&make_str_safe(&self.name)?);
         r.push_str("\", desc:\"");
-        r.push_str(&make_str_safe(&self.desc));
+        r.push_str(&make_str_safe(&self.desc)?);
         r.push_str("\", color:\"");
-        r.push_str(&make_str_safe(&self.color));
+        r.push_str(&make_str_css_safe(&self.color)?);
         r.push_str("\", children:[");
         r.push_str(
             &self
@@ -543,10 +545,11 @@ impl ToString for NodeJsonData {
                 .join(","),
         );
         r.push_str("]}");
-        r
+        Ok(r)
     }
 }
 
+// Variables that shouldn't be overwritten between cells should have ___js___ as suffix.
 const HTML_HEADER: &str = r#"
   <script>
     var expandClonestoggle___js___ = true;"#;
