@@ -590,9 +590,7 @@ pub fn remove_axes(circuit: &CircuitRc, axes: &HashSet<usize>) -> Option<Circuit
         ),
         Circuit::Rearrange(rearrange) => {
             // canonicalize first to make input 1s (), allowing us to remove them from output
-            let spec = rearrange
-                .conform_to_input_shape_spec(false)
-                .canonicalize(true);
+            let spec = rearrange.conform_to_input_shape_spec().canonicalize(true);
             Some(Rearrange::nrc_elim_identity(
                 rearrange.node.clone(),
                 spec.filter_out_axes(axes).unwrap(),
@@ -612,7 +610,7 @@ pub fn remove_axes(circuit: &CircuitRc, axes: &HashSet<usize>) -> Option<Circuit
                     // wow this chained construction is long, should change at some point
                     RearrangeSpec::ident_usize(circuit.info().rank())
                         .unwrap()
-                        .conform_to_input_shape(&circuit.info().shape, false)
+                        .conform_to_input_shape(&circuit.info().shape)
                         .unwrap()
                         .canonicalize(true)
                         .filter_out_axes(axes)
@@ -1166,8 +1164,8 @@ pub fn index_split_axes(node: &Index, top_axes: std::collections::HashSet<usize>
 pub fn rearrange_fuse(node: &Rearrange) -> Option<Rearrange> {
     node.node.as_rearrange().and_then(|inner| {
         RearrangeSpec::fuse(
-            &inner.conform_to_input_shape_spec(false),
-            &node.conform_to_input_shape_spec(false),
+            &inner.conform_to_input_shape_spec(),
+            &node.conform_to_input_shape_spec(),
         )
         .map(|spec| {
             Rearrange::new(
