@@ -83,11 +83,11 @@ pub fn batch_to_concat(circuit: CircuitRc, axis: usize, num_batches: usize) -> R
     if input_specs.is_empty() {
         bail!(BatchError::AxisOriginatesTooHigh {});
     }
-    let require_all_args = false; // this really should be true - get_axis_leaves doesn't get leaves!
+    let check_all_args_present = false; // this really should be true - get_axis_leaves doesn't get leaves!
     let module_spec = ModuleSpec::new_extract(
         circuit.clone(),
         input_specs.clone(),
-        require_all_args,
+        check_all_args_present,
         false,
     )
     .unwrap();
@@ -172,13 +172,13 @@ pub fn batch_to_concat(circuit: CircuitRc, axis: usize, num_batches: usize) -> R
 #[base_exception(PyValueError)]
 #[derive(Error, Debug, Clone)]
 pub enum BatchError {
-    #[error("num_batches {num_batches} doesn't divide length {l}")]
+    #[error("num_batches {num_batches} doesn't divide length {l} ({e_name})")]
     NumberDoesntDivide { l: usize, num_batches: usize },
 
-    #[error("Would need to batch multiple axes, only supports one")]
+    #[error("Would need to batch multiple axes, only supports one ({e_name})")]
     RequiresMultipleAxes {},
 
-    #[error("Batching axis originates too high")]
+    #[error("Batching axis originates too high ({e_name})")]
     AxisOriginatesTooHigh {},
 }
 
