@@ -7,18 +7,24 @@ use std::{
 };
 
 use anyhow::{bail, Result};
-use circuit_base::CircuitRc;
+use circuit_base::{
+    opaque_iterative_matcher::{
+        Update as BaseUpdate, UpdatedIterativeMatcher as BaseUpdatedIterativeMatcher,
+    },
+    CircuitNode, CircuitRc,
+};
 use macro_rules_attribute::apply;
-use pyo3::{prelude::*, AsPyPointer};
+use pyo3::{exceptions::PyValueError, prelude::*, AsPyPointer};
 use rr_util::{
     eq_by_big_hash::EqByBigHash,
-    impl_both_by_big_hash, setup_callable, simple_default, simple_from,
+    impl_both_by_big_hash, python_error_exception, setup_callable, simple_default, simple_from,
     tensor_util::Slice,
     util::{
         arc_ref_clone, arc_unwrap_or_clone, as_sorted, flip_op_result, split_first_take,
         EmptySingleMany as ESM, HashBytes,
     },
 };
+use thiserror::Error;
 use uuid::uuid;
 
 use crate::{
@@ -30,6 +36,7 @@ use crate::{
 #[derive(Clone, FromPyObject)]
 pub enum IterativeMatcherFromPy {
     BaseMatch(MatcherFromPyBase),
+    // Finished(Finished),
     IterativeMatcher(IterativeMatcher),
     #[pyo3(transparent)]
     PyFunc(PyObject),
@@ -38,7 +45,11 @@ pub enum IterativeMatcherFromPy {
 #[derive(Clone, Debug)]
 pub enum IterativeMatcherData {
     Match(Matcher),
+    Term(bool),
     Filter(FilterIterativeMatcher),
+    Children(ChildrenMatcher),
+    SpecCircuit(IterativeMatcherRc),
+    FilterModuleSpec(IterativeMatcherRc),
     Chains(HashSet<ChainItem>),
     Raw(RawIterativeMatcher),
     PyFunc(PyObject),
@@ -69,6 +80,9 @@ impl From<IterativeMatcherFromPy> for IterativeMatcher {
     fn from(m: IterativeMatcherFromPy) -> Self {
         match m {
             IterativeMatcherFromPy::BaseMatch(x) => x.into(),
+
+            // TODO: maybe Finished should be convertible to Term
+            // IterativeMatcherFromPy::Finished(_) => IterativeMatcherData::Term.into(),
             IterativeMatcherFromPy::IterativeMatcher(x) => x.into(),
             // we intentionally do a matcher here as the default - if users want
             // a IterativeMatcher pyfunc, they can explicitly use IterativeMatcher
@@ -85,6 +99,15 @@ pub struct ChainItem {
     hash: HashBytes,
 }
 
+#[derive(Clone, Debug, Eq, PartialEq, Ord, PartialOrd, Hash)]
+pub enum ChainItemPerChild {
+    Many(Vec<Option<ChainItem>>),
+    Single(ChainItem),
+}
+
+simple_from!(|x: ChainItem| -> ChainItemPerChild { ChainItemPerChild::Single(x) });
+simple_from!(|x: Vec<Option<ChainItem>>| -> ChainItemPerChild { ChainItemPerChild::Many(x) });
+
 impl EqByBigHash for ChainItem {
     fn hash(&self) -> HashBytes {
         self.hash
@@ -127,7 +150,11 @@ impl IterativeMatcherData {
     fn uuid(&self) -> [u8; 16] {
         match self {
             Self::Match(_) => uuid!("f1d5e8ec-cba0-496f-883e-78dd5cdc3a49"),
+            Self::Term(_) => uuid!("365bc6b4-cbc6-4d82-8e14-7505e1229eec"),
             Self::Filter(_) => uuid!("bcc5ccaa-afbe-414d-88b4-b8eac8c93ece"),
+            Self::Children(_) => uuid!("0081e9c9-7fdb-43e1-900e-f1adde28f23d"),
+            Self::SpecCircuit(_) => uuid!("b728a8c3-0493-4ee5-96c2-71e4900fc0f8"),
+            Self::FilterModuleSpec(_) => uuid!("b728a8c3-0493-4ee5-96c2-71e4900fc0f8"),
             Self::Chains(_) => uuid!("958d03ed-7a1a-4ea9-8dd4-d4a8a68feecb"),
             Self::Raw(_) => uuid!("5838ac96-0f48-4cdb-874f-d4f68ce3a52b"),
             Self::PyFunc(_) => uuid!("d3afc3c0-5c86-46df-9e41-7944caedd901"),
@@ -157,6 +184,18 @@ impl IterativeMatcherData {
                 hasher.update(&term_early_at.hash());
                 hasher.update(&depth.to_le_bytes());
             }
+            Self::Children(ChildrenMatcher {
+                iterative_matcher,
+                child_numbers,
+            }) => {
+                hasher.update(&iterative_matcher.hash);
+                for num in child_numbers {
+                    hasher.update(&num.to_le_bytes());
+                }
+            }
+            Self::SpecCircuit(matcher) | Self::FilterModuleSpec(matcher) => {
+                hasher.update(&matcher.hash);
+            }
             Self::Chains(chains) => {
                 for chain in as_sorted(chains) {
                     hasher.update(&chain.hash);
@@ -168,12 +207,14 @@ impl IterativeMatcherData {
             Self::PyFunc(x) => {
                 hasher.update(&(x.as_ptr() as usize).to_le_bytes());
             }
+            &Self::Term(next) => {
+                hasher.update(&[next as u8]);
+            }
         }
     }
 }
 
 /// Helper with some basic rules you may want to use to control your node matching iterations.
-/// TODO: include docs in py + other places as needed.
 #[derive(Clone, Debug)]
 pub struct FilterIterativeMatcher {
     pub iterative_matcher: IterativeMatcherRc,
@@ -248,8 +289,7 @@ impl FilterIterativeMatcher {
         let after_end = with_dist_of_end(0);
         if after_end {
             return Ok(IterateMatchResults {
-                updated: None,
-                finished: true,
+                updated: Some(None.into()),
                 found: false,
             });
         }
@@ -258,32 +298,30 @@ impl FilterIterativeMatcher {
         if before_start {
             return Ok(IterateMatchResults {
                 updated: Some(
-                    IterativeMatcherData::Filter(FilterIterativeMatcher {
-                        depth: self.depth + 1,
-                        ..self.clone()
-                    })
+                    Some(
+                        IterativeMatcherData::Filter(FilterIterativeMatcher {
+                            depth: self.depth + 1,
+                            ..self.clone()
+                        })
+                        .into(),
+                    )
                     .into(),
                 ),
-                finished: false,
                 found: false,
             });
         }
 
-        let IterateMatchResults {
-            updated,
-            finished,
-            found,
-        } = self.iterative_matcher.match_iterate(circuit.clone())?;
+        let IterateMatchResults { updated, found } =
+            self.iterative_matcher.match_iterate(circuit.clone())?;
 
         let reached_end = with_dist_of_end(1);
-        if finished
+        if all_finished(&updated)
             || (found && self.term_if_matches)
             || reached_end
             || self.term_early_at.call(circuit)?
         {
             return Ok(IterateMatchResults {
-                updated: None,
-                finished: true,
+                updated: Some(None.into()),
                 found,
             });
         }
@@ -293,11 +331,15 @@ impl FilterIterativeMatcher {
         } else {
             self.depth
         };
-        let updated = (self.end_depth.is_some() || updated.is_some()).then(|| {
-            let new = updated
-                .map(Into::into)
-                .unwrap_or(self.iterative_matcher.clone());
 
+        let updated = if self.end_depth.is_some() {
+            // we can't keep same if we need to update end_depth
+            Some(updated.unwrap_or(Some(self.iterative_matcher.clone()).into()))
+        } else {
+            updated
+        };
+
+        let updated = map_updated(updated, |new| {
             IterativeMatcherData::Filter(FilterIterativeMatcher {
                 iterative_matcher: new,
                 depth: new_depth,
@@ -306,21 +348,177 @@ impl FilterIterativeMatcher {
             .into()
         });
 
-        Ok(IterateMatchResults {
-            updated,
-            finished,
-            found,
+        Ok(IterateMatchResults { updated, found })
+    }
+}
+
+#[derive(Clone, Debug)]
+pub struct ChildrenMatcher {
+    pub iterative_matcher: IterativeMatcherRc,
+    pub child_numbers: HashSet<usize>,
+}
+
+impl ChildrenMatcher {
+    pub fn new(
+        iterative_matcher: IterativeMatcherRc,
+        child_numbers: HashSet<usize>,
+    ) -> Result<Self> {
+        Ok(Self {
+            iterative_matcher,
+            child_numbers,
         })
     }
+
+    pub fn check_num_children(&self, num_children: usize) -> Result<()> {
+        for n in &self.child_numbers {
+            if n >= &num_children {
+                bail!(IterativeMatcherError::ChildNumbersOutOfBounds {
+                    child_numbers: self.child_numbers.clone(),
+                    num_children
+                });
+            }
+        }
+        Ok(())
+    }
+
+    pub fn match_iterate(&self, circuit: CircuitRc) -> Result<IterateMatchResults> {
+        let num_children = circuit.num_children();
+        let IterateMatchResults { updated, found } =
+            self.iterative_matcher.match_iterate(circuit)?;
+
+        let updated_to_children_matcher = |iterative_matcher| {
+            IterativeMatcherData::Children(Self {
+                iterative_matcher,
+                child_numbers: self.child_numbers.clone(),
+            })
+            .into()
+        };
+
+        let out = if found {
+            self.check_num_children(num_children)?;
+
+            let per_child = per_child(updated, self.iterative_matcher.clone(), num_children);
+            assert_eq!(per_child.len(), num_children);
+            let updated = per_child
+                .into_iter()
+                .enumerate()
+                .map(|(i, x)| {
+                    Some(
+                        IterativeMatcher::special_case_any(
+                            x.map(updated_to_children_matcher)
+                                .into_iter()
+                                .chain(
+                                    (self.child_numbers.contains(&i))
+                                        .then(|| IterativeMatcher::term(true).into()),
+                                )
+                                .collect(),
+                        )
+                        .rc(),
+                    )
+                    .into()
+                })
+                .collect::<Vec<Update>>()
+                .into();
+
+            IterateMatchResults {
+                updated: Some(updated),
+                found: false,
+            }
+        } else {
+            IterateMatchResults {
+                updated: map_updated(updated, updated_to_children_matcher),
+                found: false,
+            }
+        };
+        Ok(out)
+    }
+}
+
+impl circuit_base::opaque_iterative_matcher::HasTerm for IterativeMatcherRc {
+    fn term() -> Self {
+        IterativeMatcher::term(false).into()
+    }
+}
+
+pub type Update = BaseUpdate<IterativeMatcherRc>;
+pub type UpdatedIterativeMatcher = BaseUpdatedIterativeMatcher<IterativeMatcherRc>;
+
+pub fn require_single(x: UpdatedIterativeMatcher) -> Result<Update> {
+    match x {
+        UpdatedIterativeMatcher::Many(matchers) => {
+            bail!(IterativeMatcherError::OperationDoesntSupportArgPerChild { matchers })
+        }
+        UpdatedIterativeMatcher::Single(x) => Ok(x),
+    }
+}
+
+pub fn per_child(
+    updated: Option<UpdatedIterativeMatcher>,
+    matcher: IterativeMatcherRc,
+    num_children: usize,
+) -> Vec<Option<IterativeMatcherRc>> {
+    updated
+        .unwrap_or(Some(matcher).into())
+        .per_child(num_children)
+}
+
+pub fn per_child_with_term(
+    updated: Option<UpdatedIterativeMatcher>,
+    matcher: IterativeMatcherRc,
+    num_children: usize,
+) -> Vec<IterativeMatcherRc> {
+    per_child(updated, matcher, num_children)
+        .into_iter()
+        .map(|x| x.unwrap_or_else(|| IterativeMatcher::term(false).into()))
+        .collect()
+}
+
+pub fn function_per_child_op(
+    updated: Option<UpdatedIterativeMatcher>,
+    matcher: IterativeMatcherRc,
+    circuit: CircuitRc,
+    mut func: impl FnMut(CircuitRc, IterativeMatcherRc) -> Result<CircuitRc>,
+) -> Result<Option<CircuitRc>> {
+    if all_finished(&updated) {
+        // special casing this isn't strictly required
+        return Ok(None);
+    }
+    let mut new_matchers = per_child(updated, matcher, circuit.num_children());
+    circuit
+        .map_children_enumerate(|i, c| {
+            std::mem::take(&mut new_matchers[i])
+                .map(|new_matcher| func(c.clone(), new_matcher))
+                .unwrap_or(Ok(c))
+        })
+        .map(Some)
+}
+
+pub fn function_per_child(
+    updated: Option<UpdatedIterativeMatcher>,
+    matcher: IterativeMatcherRc,
+    circuit: CircuitRc,
+    func: impl FnMut(CircuitRc, IterativeMatcherRc) -> Result<CircuitRc>,
+) -> Result<CircuitRc> {
+    function_per_child_op(updated, matcher, circuit.clone(), func).map(|x| x.unwrap_or(circuit))
+}
+
+pub fn map_updated(
+    x: Option<UpdatedIterativeMatcher>,
+    f: impl FnMut(IterativeMatcherRc) -> IterativeMatcherRc,
+) -> Option<UpdatedIterativeMatcher> {
+    x.map(|x| x.map_updated(f))
+}
+
+pub fn all_finished(x: &Option<UpdatedIterativeMatcher>) -> bool {
+    x.as_ref().map(|x| x.all_finished()).unwrap_or(false)
 }
 
 #[pyclass]
 #[derive(Clone, Debug)]
 pub struct IterateMatchResults {
+    /// here None, is 'use same'
     #[pyo3(set, get)]
-    pub updated: Option<IterativeMatcherRc>,
-    #[pyo3(set, get)]
-    pub finished: bool,
+    pub updated: Option<UpdatedIterativeMatcher>,
     #[pyo3(set, get)]
     pub found: bool,
 }
@@ -328,37 +526,36 @@ pub struct IterateMatchResults {
 #[pymethods]
 impl IterateMatchResults {
     #[new]
-    #[args(updated = "None", finished = "false", found = "false")]
-    fn new(updated: Option<IterativeMatcherRc>, finished: bool, found: bool) -> Self {
+    #[args(updated = "None", found = "false")]
+    fn new(updated: Option<UpdatedIterativeMatcher>, found: bool) -> Self {
+        Self { updated, found }
+    }
+
+    #[staticmethod]
+    #[args(found = "false")]
+    pub fn new_finished(found: bool) -> Self {
         Self {
-            updated,
-            finished,
+            updated: Some(None.into()),
             found,
         }
     }
 
-    fn to_tup(&self) -> (Option<IterativeMatcherRc>, bool, bool) {
-        (self.updated.clone(), self.finished, self.found)
+    fn to_tup(&self) -> (Option<UpdatedIterativeMatcher>, bool) {
+        (self.updated.clone(), self.found)
     }
 
-    #[pyo3(name = "none_if_finished")]
-    pub fn none_if_finished_py(
+    #[pyo3(name = "unwrap_or_same")]
+    pub fn unwrap_or_same_py(
         &self,
         matcher: IterativeMatcherRc,
-    ) -> (Option<IterativeMatcherRc>, bool) {
-        self.clone().none_if_finished(matcher)
+    ) -> (UpdatedIterativeMatcher, bool) {
+        self.clone().unwrap_or_same(matcher)
     }
 }
 
 impl IterateMatchResults {
-    pub fn none_if_finished(
-        self,
-        matcher: IterativeMatcherRc,
-    ) -> (Option<IterativeMatcherRc>, bool) {
-        (
-            (!self.finished).then(|| self.updated.unwrap_or(matcher)),
-            self.found,
-        )
+    pub fn unwrap_or_same(self, matcher: IterativeMatcherRc) -> (UpdatedIterativeMatcher, bool) {
+        (self.updated.unwrap_or(Some(matcher).into()), self.found)
     }
 }
 
@@ -367,10 +564,15 @@ impl IterativeMatcher {
         Self::any(vec![self.crc(), other])
     }
 
+    // TODO: this is kinda shitty, see https://github.com/redwoodresearch/unity/issues/1524
     pub fn validate_matched(&self, matched: &HashSet<CircuitRc>) -> Result<()> {
         match &self.data {
             IterativeMatcherData::Match(m) => m.validate_matched(matched),
+            IterativeMatcherData::Term(_) => Ok(()),
             IterativeMatcherData::Filter(m) => m.iterative_matcher.validate_matched(matched),
+            IterativeMatcherData::Children(_) => Ok(()), /* not sure what else you could do here */
+            IterativeMatcherData::SpecCircuit(_) => Ok(()), /* not sure what else you could do here */
+            IterativeMatcherData::FilterModuleSpec(m) => m.validate_matched(matched),
             IterativeMatcherData::Chains(m) => {
                 for chain in m {
                     chain.last().validate_matched(matched)?;
@@ -387,9 +589,10 @@ impl IterativeMatcher {
 impl IterativeMatcher {
     #[new]
     #[args(inps = "*")]
-    fn py_new(inps: Vec<IterativeMatcherRc>) -> Self {
+    // also py_new
+    fn special_case_any(inps: Vec<IterativeMatcherRc>) -> Self {
         match inps.into() {
-            ESM::Empty => MatcherData::Always(false).into(),
+            ESM::Empty => Self::term(false),
             ESM::Single(x) => arc_ref_clone(&x),
             ESM::Many(x) => Self::any(x),
         }
@@ -400,14 +603,102 @@ impl IterativeMatcher {
         MatcherData::Always(true).into()
     }
 
+    #[staticmethod]
+    #[args(match_next = "false")]
+    pub fn term(match_next: bool) -> Self {
+        IterativeMatcherData::Term(match_next).into()
+    }
+
     pub fn match_iterate(&self, circuit: CircuitRc) -> Result<IterateMatchResults> {
+        let num_children = circuit.num_children();
         let res = match &self.data {
             IterativeMatcherData::Match(m) => IterateMatchResults {
                 updated: None,
-                finished: false,
                 found: m.call(circuit)?,
             },
+            &IterativeMatcherData::Term(match_next) => {
+                IterateMatchResults::new_finished(match_next)
+            }
             IterativeMatcherData::Filter(filter) => filter.match_iterate(circuit)?,
+            IterativeMatcherData::Children(children_matcher) => {
+                children_matcher.match_iterate(circuit)?
+            }
+            IterativeMatcherData::SpecCircuit(module_matcher) => {
+                let is_module = circuit.is_module();
+                let IterateMatchResults { updated, found } =
+                    module_matcher.match_iterate(circuit)?;
+                let updated_to_spec_circuit_matcher =
+                    |iterative_matcher| IterativeMatcherData::SpecCircuit(iterative_matcher).into();
+                if found {
+                    if !is_module {
+                        bail!("not module!");
+                    }
+
+                    let per_child = per_child(updated, module_matcher.clone(), num_children);
+
+                    let updated = per_child
+                        .into_iter()
+                        .enumerate()
+                        .map(|(i, x)| {
+                            Some(
+                                IterativeMatcher::special_case_any(
+                                    x.map(updated_to_spec_circuit_matcher)
+                                        .into_iter()
+                                        .chain(
+                                            (i == 0) // spec circuit is child 0 on Module
+                                                .then(|| IterativeMatcher::term(true).into()),
+                                        )
+                                        .collect(),
+                                )
+                                .rc(),
+                            )
+                            .into()
+                        })
+                        .collect::<Vec<Update>>()
+                        .into();
+
+                    IterateMatchResults {
+                        updated: Some(updated),
+                        found: false,
+                    }
+                } else {
+                    IterateMatchResults {
+                        updated: map_updated(updated, updated_to_spec_circuit_matcher),
+                        found: false,
+                    }
+                }
+            }
+            IterativeMatcherData::FilterModuleSpec(matcher) => {
+                let is_module = circuit.is_module();
+                let IterateMatchResults { updated, found } = matcher.match_iterate(circuit)?;
+                if is_module {
+                    let per_child = per_child(updated, matcher.clone(), num_children);
+                    let updated = per_child
+                        .into_iter()
+                        .enumerate()
+                        .map(|(i, x)| {
+                            x.and_then(|x| {
+                                (i != 0 && (i - 1) % 2 != 0)
+                                    .then(|| IterativeMatcherData::FilterModuleSpec(x).into())
+                            })
+                            .into()
+                        })
+                        .collect::<Vec<Update>>()
+                        .into();
+
+                    IterateMatchResults {
+                        updated: Some(updated),
+                        found,
+                    }
+                } else {
+                    IterateMatchResults {
+                        updated: map_updated(updated, |x| {
+                            IterativeMatcherData::FilterModuleSpec(x).into()
+                        }),
+                        found,
+                    }
+                }
+            }
             IterativeMatcherData::Chains(chains) => {
                 /// avoid some hashing and some copies - probably overkill
                 #[derive(Clone, Copy)]
@@ -433,26 +724,42 @@ impl IterativeMatcher {
                             Self::Slices { rest, .. } => rest,
                         }
                     }
+
+                    fn clone_chain(&'a self) -> ChainItem {
+                        match self {
+                            &Self::Chain(x) => x.clone(),
+                            Self::Slices { first, rest } => {
+                                ChainItem::new((*first).clone(), rest.to_vec())
+                            }
+                        }
+                    }
+
+                    fn with_first(&'a self, first: IterativeMatcherRc) -> ChainItem {
+                        ChainItem::new(first, self.rest().to_vec()).into()
+                    }
                 }
 
                 fn run_item(
                     chain: MaybeChain<'_>,
                     circuit: &CircuitRc,
-                    new_items: &mut HashSet<ChainItem>,
+                    new_items: &mut HashSet<ChainItemPerChild>,
                 ) -> Result<bool> {
-                    let IterateMatchResults {
-                        updated,
-                        finished,
-                        found,
-                    } = chain.first().match_iterate(circuit.clone())?;
+                    let IterateMatchResults { updated, found } =
+                        chain.first().match_iterate(circuit.clone())?;
 
-                    if !finished {
+                    if !all_finished(&updated) {
+                        use BaseUpdatedIterativeMatcher::*;
                         let new_chain = match (updated, chain) {
-                            (None, MaybeChain::Chain(chain)) => chain.clone(),
-                            (Some(x), chain) => ChainItem::new(x, chain.rest().to_vec()),
-                            (None, MaybeChain::Slices { first, .. }) => {
-                                ChainItem::new(first.clone(), chain.rest().to_vec())
+                            (None, _) => chain.clone_chain().into(),
+                            (Some(Single(BaseUpdate(None))), _) => unreachable!(),
+                            (Some(Single(BaseUpdate(Some(x)))), chain) => {
+                                chain.with_first(x).into()
                             }
+                            (Some(Many(items)), chain) => items
+                                .into_iter()
+                                .map(|x| x.0.map(|x| chain.with_first(x)))
+                                .collect::<Vec<_>>()
+                                .into(),
                         };
                         new_items.insert(new_chain);
                     }
@@ -477,17 +784,64 @@ impl IterativeMatcher {
                 let mut new_items = HashSet::new();
                 let mut any_chain_finished = false;
                 for chain in chains {
-                    any_chain_finished =
-                        run_item(MaybeChain::Chain(chain), &circuit, &mut new_items)?
-                            || any_chain_finished;
+                    let new_any_finished =
+                        run_item(MaybeChain::Chain(chain), &circuit, &mut new_items)?;
+                    any_chain_finished = new_any_finished || any_chain_finished;
                 }
 
                 let finished = new_items.is_empty();
 
+                let updated = if new_items
+                    .iter()
+                    .any(|x| matches!(x, ChainItemPerChild::Many(_)))
+                {
+                    assert!(!finished);
+                    let num_children = circuit.children().count();
+                    let mut out = vec![HashSet::default(); num_children];
+                    for item in new_items {
+                        match item {
+                            ChainItemPerChild::Single(x) => {
+                                out.iter_mut().for_each(|child_chains| {
+                                    child_chains.insert(x.clone());
+                                })
+                            }
+                            ChainItemPerChild::Many(xs) => {
+                                out.iter_mut().zip(xs).for_each(|(child_chains, x)| {
+                                    if let Some(x) = x {
+                                        child_chains.insert(x);
+                                    }
+                                })
+                            }
+                        }
+                    }
+                    Some(UpdatedIterativeMatcher::Many(
+                        out.into_iter()
+                            .map(|chains| {
+                                (!chains.is_empty())
+                                    .then(|| IterativeMatcherData::Chains(chains).into())
+                                    .into()
+                            })
+                            .collect(),
+                    ))
+                } else {
+                    let new_items = new_items
+                        .into_iter()
+                        .map(|x| match x {
+                            ChainItemPerChild::Single(x) => x,
+                            ChainItemPerChild::Many(_) => unreachable!(),
+                        })
+                        .collect();
+                    (&new_items != chains).then(|| {
+                        UpdatedIterativeMatcher::Single(
+                            (!finished)
+                                .then(|| IterativeMatcherData::Chains(new_items).into())
+                                .into(),
+                        )
+                    })
+                };
+
                 IterateMatchResults {
-                    finished,
-                    updated: (!finished && &new_items != chains)
-                        .then(|| IterativeMatcherData::Chains(new_items).into()),
+                    updated,
                     found: any_chain_finished,
                 }
             }
@@ -497,6 +851,18 @@ impl IterativeMatcher {
             }
         };
 
+        if let Some(UpdatedIterativeMatcher::Many(items)) = &res.updated {
+            // checking here allows us to assume valid in other places!
+            if items.len() != num_children {
+                bail!(IterativeMatcherError::NumUpdatedMatchersNEQToNumChildren {
+                    num_updated_matchers: items.len(),
+                    num_children,
+                    updated_matchers: items.clone(),
+                    from_matcher: self.crc()
+                });
+            }
+        }
+
         Ok(res)
     }
 
@@ -519,13 +885,13 @@ impl IterativeMatcher {
 
     #[staticmethod]
     #[args(first, rest = "*")]
-    pub fn mk_chain(first: IterativeMatcherRc, rest: Vec<IterativeMatcherRc>) -> Self {
+    pub fn new_chain(first: IterativeMatcherRc, rest: Vec<IterativeMatcherRc>) -> Self {
         first.chain(rest)
     }
 
     #[staticmethod]
     #[args(chains = "*")]
-    pub fn mk_chain_many(chains: Vec<Vec<IterativeMatcherRc>>) -> Result<Self> {
+    pub fn new_chain_many(chains: Vec<Vec<IterativeMatcherRc>>) -> Result<Self> {
         Ok(IterativeMatcherData::Chains(
             chains
                 .into_iter()
@@ -539,13 +905,26 @@ impl IterativeMatcher {
     }
 
     #[staticmethod]
-    #[pyo3(name = "mk_func")]
-    pub(super) fn mk_func_py(f: PyObject) -> Self {
+    pub fn new_children_matcher(
+        first_match: IterativeMatcherRc,
+        child_numbers: HashSet<usize>,
+    ) -> Self {
+        first_match.children_matcher(child_numbers)
+    }
+
+    #[staticmethod]
+    pub fn new_spec_circuit_matcher(module_matcher: IterativeMatcherRc) -> Self {
+        module_matcher.spec_circuit_matcher()
+    }
+
+    #[staticmethod]
+    #[pyo3(name = "new_func")]
+    pub(super) fn new_func_py(f: PyObject) -> Self {
         IterativeMatcherData::PyFunc(f).into()
     }
 
     #[args(others = "*")]
-    pub fn mk_or(&self, others: Vec<IterativeMatcherRc>) -> Self {
+    pub fn new_or(&self, others: Vec<IterativeMatcherRc>) -> Self {
         Self::any([self.clone().into()].into_iter().chain(others).collect())
     }
 
@@ -673,6 +1052,27 @@ impl IterativeMatcher {
         .into()
     }
 
+    pub fn children_matcher(&self, child_numbers: HashSet<usize>) -> IterativeMatcher {
+        IterativeMatcherData::Children(ChildrenMatcher {
+            iterative_matcher: self_.clone().rc(),
+            child_numbers,
+        })
+        .into()
+    }
+
+    pub fn spec_circuit_matcher(&self) -> IterativeMatcher {
+        IterativeMatcherData::SpecCircuit(self_.clone().rc()).into()
+    }
+
+    #[args(enable = "true")]
+    pub fn filter_module_spec(&self, enable: bool) -> IterativeMatcher {
+        if enable {
+            IterativeMatcherData::FilterModuleSpec(self_.clone().rc()).into()
+        } else {
+            self_.clone()
+        }
+    }
+
     #[args(fancy_validate = "Getter::default().default_fancy_validate")]
     pub fn get(&self, circuit: CircuitRc, fancy_validate: bool) -> Result<HashSet<CircuitRc>> {
         Getter::default().get(circuit, self_.crc(), Some(fancy_validate))
@@ -773,3 +1173,28 @@ impl IterativeMatcher {
         )
     }
 }
+
+#[apply(python_error_exception)]
+#[base_error_name(IterativeMatcher)]
+#[base_exception(PyValueError)]
+#[derive(Error, Debug, Clone)]
+pub enum IterativeMatcherError {
+    #[error(
+        "num_updated_matchers={num_updated_matchers} != num_children={num_children}\nupdated_matchers={updated_matchers:?}, from_matcher={from_matcher:?}\n({e_name})"
+    )]
+    NumUpdatedMatchersNEQToNumChildren {
+        num_updated_matchers: usize,
+        num_children: usize,
+        updated_matchers: Vec<Update>,
+        from_matcher: IterativeMatcherRc,
+    },
+
+    #[error("operation doesn't support per child matching matchers={matchers:?}\n({e_name})")]
+    OperationDoesntSupportArgPerChild { matchers: Vec<Update> },
+
+    #[error("some child_numbers={child_numbers:?} >= num_children={num_children} ({e_name})")]
+    ChildNumbersOutOfBounds {
+        child_numbers: HashSet<usize>,
+        num_children: usize,
+    },
+}
