@@ -5,12 +5,12 @@ pub mod iterative_matcher;
 pub mod library;
 pub mod matcher;
 pub mod operations;
-pub mod sample_transform;
+pub mod sampler;
 pub mod transform;
 
 pub use iterative_matcher::{
     IterateMatchResults, IterativeMatcher, IterativeMatcherData, IterativeMatcherRc,
 };
 pub use matcher::{Matcher, MatcherData, MatcherFromPy, MatcherFromPyBase, MatcherRc};
-pub use operations::{Batcher, BoundGetter, BoundUpdater, Getter, Updater};
+pub use operations::{BoundGetter, BoundUpdater, Expander, Getter, Updater};
 pub use transform::{Transform, TransformData, TransformRc};
