@@ -380,7 +380,7 @@ impl BoundGetter {
 
 #[pyclass]
 #[derive(Debug, Clone)]
-pub struct Batcher {
+pub struct Expander {
     /// Note: we don't currently cache these transforms individually. We could
     /// do this.
     #[pyo3(get)]
@@ -402,7 +402,7 @@ pub struct Batcher {
     pub(super) validation_getter: Getter,
 }
 
-impl Default for Batcher {
+impl Default for Expander {
     fn default() -> Self {
         Self {
             replacements: Vec::new(),
@@ -417,21 +417,21 @@ impl Default for Batcher {
 }
 
 #[pymethods]
-impl Batcher {
+impl Expander {
     #[new]
     #[args(
-        batchers = "*",
-        ban_multiple_matches_on_node = "Batcher::default().ban_multiple_matches_on_node",
-        default_fancy_validate = "Batcher::default().default_fancy_validate",
+        expanders = "*",
+        ban_multiple_matches_on_node = "Expander::default().ban_multiple_matches_on_node",
+        default_fancy_validate = "Expander::default().default_fancy_validate",
         suffix = "None"
     )]
     pub fn new(
-        batchers: Vec<(IterativeMatcherRc, TransformRc)>,
+        expanders: Vec<(IterativeMatcherRc, TransformRc)>,
         ban_multiple_matches_on_node: bool,
         default_fancy_validate: bool,
         suffix: Option<String>,
     ) -> Self {
-        let (matchers, replacements) = batchers
+        let (matchers, replacements) = expanders
             .into_iter()
             .map(|(a, b)| (a.into(), b.into()))
             .unzip();
@@ -465,7 +465,7 @@ impl Batcher {
     }
 }
 
-impl Batcher {
+impl Expander {
     #[apply(cached_method)]
     #[self_id(self_)]
     #[key((circuit.info().hash, matchers.clone()))]
