@@ -161,7 +161,7 @@ pub fn make_diagonal(
 pub fn einops_repeat(
     tensor: &Tensor,
     op: String,
-    sizes: impl IntoIterator<Item = (String, u64)>,
+    sizes: impl IntoIterator<Item = (String, usize)>,
 ) -> Result<Tensor> {
     PY_EINOPS
         .repeat
@@ -571,6 +571,10 @@ pub static PY_CIRCUIT_ITEMS: GILLazyPy<PyCircuitItems> = GILLazyPy::new_py(|py|
     }
 });
 
+pub static SELF_MODULE: GILLazyPy<Py<PyModule>> = GILLazyPy::new_py(|py| {
+    PyModule::import(py, "rust_circuit").unwrap().into() // obviously a bit cursed : (
+});
+
 #[pyclass]
 #[derive(Copy, Clone)]
 pub struct NotSet;
