@@ -1,14 +1,15 @@
 use std::{
-    fmt::Debug,
+    fmt::{self, Debug},
     hash,
     iter::{self, zip},
+    str::FromStr,
 };
 
 use anyhow::{bail, Context, Result};
 use itertools::{izip, Itertools};
 use macro_rules_attribute::apply;
 use once_cell::sync::Lazy;
-use pyo3::{exceptions::PyValueError, prelude::*, types::IntoPyDict};
+use pyo3::{exceptions::PyValueError, prelude::*};
 use regex::Regex;
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use smallvec::SmallVec as Sv;
@@ -17,9 +18,10 @@ use uuid::uuid;
 
 use crate::{
     make_single_many,
-    py_types::{PyOpAtAxes, PY_CIRCUIT_ITEMS},
+    py_types::{einops_repeat, PyOpAtAxes, Tensor, PY_CIRCUIT_ITEMS},
     python_error_exception, sv,
-    tensor_util::{check_canon_idxs, Shape},
+    symbolic_size::{SymbolicSizeProduct, SIZE_PROD_MATCH},
+    tensor_util::{check_canon_idxs, PyParseError, Shape},
     util::{
         dict_to_list, filter_out_idx, is_unique, vec_map_insert, AxisInt, EinsumAxes, HashBytes,
         ALPHABET,
@@ -47,14 +49,14 @@ impl OpSize {
     pub fn is_none(&self) -> bool {
         self.0 >> Self::SHIFT != 0
     }
-    pub fn unwrap(self) -> u64 {
+    pub fn unwrap(self) -> usize {
         assert!(self.is_some());
-        self.0
+        self.0 as usize
     }
 }
 
 impl Debug for OpSize {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         self.t().fmt(f)
     }
 }
@@ -104,7 +106,7 @@ pub type RInnerInts = Sv<[AxisInt; 8]>;
 pub type RInts = Sv<[RInnerInts; 8]>;
 
 #[pyclass]
-#[derive(Debug, Clone, PartialEq, Eq)]
+#[derive(Clone, PartialEq, Eq)]
 pub struct RearrangeSpec {
     #[pyo3(get)]
     pub input_ints: RInts,
@@ -132,6 +134,111 @@ pub fn check_permutation<T: Eq + hash::Hash + Clone + Into<usize>>(perm: &[T]) -
     Ok(())
 }
 
+impl fmt::Display for RearrangeSpec {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        // plausible that default display shouldn't show_size_if_present, but debug should. (idk)
+        f.write_str(&self.to_string_impl(true, true, true))
+    }
+}
+impl fmt::Debug for RearrangeSpec {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        fmt::Display::fmt(self, f)
+    }
+}
+
+impl FromStr for RearrangeSpec {
+    type Err = anyhow::Error;
+    fn from_str(string: &str) -> Result<Self, Self::Err> {
+        const ALPHA_NON_NUMERIC: &str = "[a-zA-Z]";
+        static RE_TOP_LEVEL: Lazy<Regex> = Lazy::new(|| {
+            Regex::new(&format!(
+                r"({p})|({a}\w*(?::{p})?)|\(\s*((?:({a}\w*(?::{p})?|{p})\s*)*)\)",
+                a = ALPHA_NON_NUMERIC,
+                p = *SIZE_PROD_MATCH
+            ))
+            .unwrap()
+        });
+
+        let (input, output) =
+            if let [input, output] = string.split("->").map(str::trim).collect::<Vec<_>>()[..] {
+                (input, output)
+            } else {
+                bail!(RearrangeParseError::ArrowIssue {
+                    string: string.to_owned(),
+                });
+            };
+
+        let mut sizes: OpShape = sv![];
+        let mut string_to_int: HashMap<String, u8> = HashMap::default();
+        let new_int = |s: usize| {
+            assert!(s > 0);
+            if s - 1 > (u8::MAX as usize) {
+                bail!(RearrangeParseError::TooManyAxes {
+                    string: string.to_owned()
+                });
+            }
+            Ok((s - 1) as u8)
+        };
+        let mut f = |s: &str| -> Result<RInts> {
+            let handle_int = |num, sizes: &mut OpShape| {
+                sizes.push(Some(num).into());
+                new_int(sizes.len())
+            };
+            let parse_num = |s: &str| {
+                SymbolicSizeProduct::parse_to_usize(s).context("rearrange num didn't parse")
+            };
+            let mut handle_string_nosize = |s: String, sizes: &mut OpShape| {
+                if let Some(old) = string_to_int.get(&s) {
+                    Ok(*old)
+                } else {
+                    sizes.push(None.into());
+                    let new_int = new_int(sizes.len())?;
+                    string_to_int.insert(s.to_owned(), new_int);
+                    Ok(new_int)
+                }
+            };
+            let mut handle_string = |string: String, sizes: &mut OpShape| -> Result<u8> {
+                if let [string, num] = string.split(':').collect::<Vec<_>>()[..] {
+                    let num = parse_num(num)?;
+                    let result = handle_string_nosize(string.to_owned(), sizes)?;
+                    sizes[result as usize] = Some(num).into();
+                    Ok(result)
+                } else {
+                    handle_string_nosize(string, sizes)
+                }
+            };
+            RE_TOP_LEVEL
+                .captures_iter(s)
+                .map(|tl| {
+                    if let Some(c) = tl.get(1) {
+                        let num = parse_num(c.as_str())?;
+                        Ok(sv![handle_int(num, &mut sizes)?])
+                    } else if let Some(c) = tl.get(2) {
+                        Ok(sv![handle_string(c.as_str().to_owned(), &mut sizes)?])
+                    } else if let Some(c) = tl.get(3) {
+                        c.as_str()
+                            .split_whitespace()
+                            .filter(|z| !z.is_empty())
+                            .map(|s| {
+                                parse_num(s)
+                                    .and_then(|z| handle_int(z, &mut sizes))
+                                    .or_else(|_e| handle_string(s.to_owned(), &mut sizes))
+                            })
+                            .collect::<Result<_>>()
+                    } else {
+                        bail!(RearrangeParseError::FailedToMatchRegex {
+                            string: string.to_owned(),
+                        });
+                    }
+                })
+                .collect()
+        };
+
+        Self::new(f(input)?, f(output)?, sizes)
+            .context("failed to construct rearrange spec after parsing")
+    }
+}
+
 /// RearrangeSpec encodes the same thing as a python Einops.repeat string https://einops.rocks/api/repeat/
 ///
 /// Dimension names are encoded as integers, so the einops operation
@@ -280,7 +387,7 @@ impl RearrangeSpec {
                     py,
                     (self
                         .canonicalize(true)
-                        .conform_to_input_shape(&input_shape, true)
+                        .conform_to_input_shape(&input_shape)
                         .unwrap()
                         .fill_empty_ints_allow_invalid(),),
                     None,
@@ -379,164 +486,78 @@ impl RearrangeSpec {
         self.check_valid().is_ok()
     }
 
-    pub fn to_einops_string(&self, size_on_letter: bool) -> String {
-        let ints_in_output = self.ints_in_out();
-        let ints_needing_sizes = self.ints_needing_sizes();
-        let input_letters = self
-            .input_ints
-            .iter()
-            .map(|one_axis_ints| {
-                let one_axis_letters = one_axis_ints
-                    .iter()
-                    .map(|i| {
-                        if ints_in_output.contains(i) || self.int_sizes[*i as usize].is_none() {
-                            if size_on_letter
-                                && ints_needing_sizes.contains(i)
-                                && self.int_sizes[*i as usize].is_some()
-                            {
-                                format!(
-                                    "{}:{}",
-                                    ALPHABET[*i as usize].to_owned(),
-                                    Option::<usize>::from(self.int_sizes[*i as usize]).unwrap()
-                                )
+    pub fn to_einops_string(&self) -> String {
+        self.to_string_impl(false, false, false)
+    }
+
+    #[pyo3(name = "to_string")]
+    #[args(
+        size_on_letter = "true",
+        use_symbolic = "true",
+        show_size_if_present = "true"
+    )]
+    pub fn to_string_impl(
+        &self,
+        size_on_letter: bool,
+        use_symbolic: bool,
+        show_size_if_present: bool,
+    ) -> String {
+        let ints_needing_sizes = (!show_size_if_present).then(|| self.ints_needing_sizes());
+        let ints_needing_sizes = ints_needing_sizes.as_ref();
+
+        let int_to_str = |i: usize| {
+            if use_symbolic {
+                SymbolicSizeProduct::from(i).to_string()
+            } else {
+                i.to_string()
+            }
+        };
+
+        let get_str = |ints: &RInts, ints_in_other: &HashSet<AxisInt>| {
+            ints.iter()
+                .map(|one_axis_ints| {
+                    let one_axis_letters = one_axis_ints
+                        .iter()
+                        .map(|i| {
+                            if ints_in_other.contains(i) || self.int_sizes[*i as usize].is_none() {
+                                if size_on_letter
+                                    && ints_needing_sizes.map(|x| x.contains(i)).unwrap_or(true)
+                                    && self.int_sizes[*i as usize].is_some()
+                                {
+                                    format!(
+                                        "{}:{}",
+                                        ALPHABET[*i as usize].to_owned(),
+                                        int_to_str(self.int_sizes[*i as usize].unwrap() as usize)
+                                    )
+                                } else {
+                                    ALPHABET[*i as usize].to_owned()
+                                }
                             } else {
-                                ALPHABET[*i as usize].to_owned()
+                                int_to_str(self.int_sizes[*i as usize].unwrap() as usize)
                             }
-                        } else {
-                            self.int_sizes[*i as usize].0.to_string()
-                        }
-                    })
-                    .collect::<Vec<String>>();
-                if one_axis_letters.len() == 1 {
-                    one_axis_letters[0].to_owned()
-                } else {
-                    format!("({})", one_axis_letters.join(" "))
-                }
-            })
-            .collect::<Vec<String>>()
-            .join(" ");
-        let ints_in_input = self.ints_in_inp();
-        let output_letters = self
-            .output_ints
-            .iter()
-            .map(|one_axis_ints| {
-                let one_axis_letters = one_axis_ints
-                    .iter()
-                    .map(|i| {
-                        if ints_in_input.contains(i) || self.int_sizes[*i as usize].is_none() {
-                            if size_on_letter
-                                && ints_needing_sizes.contains(i)
-                                && self.int_sizes[*i as usize].is_some()
-                            {
-                                format!(
-                                    "{}:{}",
-                                    ALPHABET[*i as usize].to_owned(),
-                                    Option::<usize>::from(self.int_sizes[*i as usize]).unwrap()
-                                )
-                            } else {
-                                ALPHABET[*i as usize].to_owned()
-                            }
-                        } else {
-                            self.int_sizes[*i as usize].0.to_string()
-                        }
-                    })
-                    .collect::<Vec<String>>();
-                if one_axis_letters.len() == 1 {
-                    one_axis_letters[0].to_owned()
-                } else {
-                    format!("({})", one_axis_letters.join(" "))
-                }
-            })
-            .collect::<Vec<String>>()
-            .join(" ");
-        input_letters + " -> " + &output_letters
+                        })
+                        .collect::<Vec<String>>();
+                    if one_axis_letters.len() == 1 {
+                        one_axis_letters[0].to_owned()
+                    } else {
+                        format!("({})", one_axis_letters.join(" "))
+                    }
+                })
+                .collect::<Vec<String>>()
+                .join(" ")
+        };
+
+        format!(
+            "{} -> {}",
+            get_str(&self.input_ints, &self.ints_in_out()),
+            get_str(&self.output_ints, &self.ints_in_inp())
+        )
     }
 
     #[staticmethod]
-    pub fn from_einops_string(string: &str) -> Result<Self> {
-        static RE_TOP_LEVEL: Lazy<Regex> = Lazy::new(|| {
-            Regex::new(r"(\d+)|(\w[\w\d]*(?::\d+)?)|\(\s*((?:(\w[\w\d]*(?::\d+)?|\d+)\s*)*)\)")
-                .unwrap()
-        });
-
-        let (input, output) =
-            if let [input, output] = string.split("->").map(str::trim).collect::<Vec<_>>()[..] {
-                (input, output)
-            } else {
-                bail!(RearrangeParseError::ArrowIssue {
-                    string: string.to_owned(),
-                });
-            };
-
-        let mut sizes: OpShape = sv![];
-        let mut string_to_int: HashMap<String, u8> = HashMap::default();
-        let new_int = |s: usize| {
-            assert!(s > 0);
-            if s - 1 > (u8::MAX as usize) {
-                bail!(RearrangeParseError::TooManyAxes {
-                    string: string.to_owned()
-                });
-            }
-            Ok((s - 1) as u8)
-        };
-        let mut f = |s: &str| -> Result<RInts> {
-            let handle_int = |num, sizes: &mut OpShape| {
-                sizes.push(Some(num).into());
-                new_int(sizes.len())
-            };
-            let mut handle_string_nosize = |s: String, sizes: &mut OpShape| {
-                if let Some(old) = string_to_int.get(&s) {
-                    Ok(*old)
-                } else {
-                    sizes.push(None.into());
-                    let new_int = new_int(sizes.len())?;
-                    string_to_int.insert(s.to_owned(), new_int);
-                    Ok(new_int)
-                }
-            };
-            let mut handle_string = |string: String, sizes: &mut OpShape| -> Result<u8> {
-                if let [string, num] = string.split(':').collect::<Vec<_>>()[..] {
-                    let num = num.parse::<usize>().context("rearrange num didn't parse")?;
-                    let result = handle_string_nosize(string.to_owned(), sizes)?;
-                    sizes[result as usize] = Some(num).into();
-                    Ok(result)
-                } else {
-                    handle_string_nosize(string, sizes)
-                }
-            };
-            RE_TOP_LEVEL
-                .captures_iter(s)
-                .map(|tl| {
-                    if let Some(c) = tl.get(1) {
-                        let num = c
-                            .as_str()
-                            .parse::<usize>()
-                            .context("rearrange num didn't parse")?;
-                        Ok(sv![handle_int(num, &mut sizes)?])
-                    } else if let Some(c) = tl.get(2) {
-                        Ok(sv![handle_string(c.as_str().to_owned(), &mut sizes)?])
-                    } else if let Some(c) = tl.get(3) {
-                        c.as_str()
-                            .split_whitespace()
-                            .filter(|z| !z.is_empty())
-                            .map(|s| {
-                                s.parse::<usize>()
-                                    .context("rearrange num didn't parse")
-                                    .and_then(|z| handle_int(z, &mut sizes))
-                                    .or_else(|_e| handle_string(s.to_owned(), &mut sizes))
-                            })
-                            .collect::<Result<_>>()
-                    } else {
-                        bail!(RearrangeParseError::FailedToMatchRegex {
-                            string: string.to_owned(),
-                        });
-                    }
-                })
-                .collect()
-        };
-
-        Self::new(f(input)?, f(output)?, sizes)
-            .context("failed to construct rearrange spec after parsing")
+    #[pyo3(name = "from_string")]
+    pub fn py_from_string(string: &str) -> Result<Self> {
+        string.parse()
     }
 
     pub fn add_batch_dims(&self, batch_rank: usize) -> Self {
@@ -557,6 +578,8 @@ impl RearrangeSpec {
         )
         .unwrap()
     }
+
+    /// does this int ever appear in an inner group with >1 element?
     pub fn ints_needing_sizes(&self) -> HashSet<AxisInt> {
         self.input_ints
             .iter()
@@ -565,36 +588,14 @@ impl RearrangeSpec {
             .copied()
             .collect()
     }
-    pub fn letter_sizes(&self) -> Vec<(String, u64)> {
-        // because to_einops_string uses numerals on all added dims, we only need letter sizes for splits
-
-        self.ints_needing_sizes()
-            .into_iter()
-            .map(|x| {
-                let size_here = self.int_sizes[x as usize];
-                assert!(size_here.is_some());
-                (ALPHABET[x as usize].to_owned(), size_here.0)
-            })
-            .collect()
-    }
-
-    pub fn to_einops_string_and_letter_sizes(&self) -> (String, Vec<(String, u64)>) {
-        (self.to_einops_string(false), self.letter_sizes())
-    }
 
     /// apply rearrange to tensor
-    pub fn apply(&self, tensor: &PyAny) -> PyResult<Py<PyAny>> {
-        let (string, letter_sizes) = self.to_einops_string_and_letter_sizes();
-        let result = Python::with_gil(|py| {
-            let einops = PyModule::import(py, "einops").unwrap();
-            let result = einops
-                .getattr("repeat")
-                .unwrap()
-                .call((tensor, string), Some(letter_sizes.into_py_dict(py)))
-                .unwrap();
-            result.to_object(py)
-        });
-        Ok(result)
+    pub fn apply(&self, tensor: Tensor) -> Result<Tensor> {
+        let (string, letter_sizes) = self
+            .conform_to_input_shape(&tensor.shape())
+            .context("conform to input shape failed in apply")?
+            .to_einops_string_and_letter_sizes();
+        einops_repeat(&tensor, string, letter_sizes)
     }
     fn __repr__(&self) -> String {
         format!("{:?}", self)
@@ -867,9 +868,19 @@ impl RearrangeSpec {
         RearrangeSpec::fuse(&inner, &outer)
     }
     #[pyo3(name = "conform_to_input_shape")]
-    fn conform_to_input_shape_py(&self, shape: Shape, coerce: bool) -> Result<RearrangeSpec> {
-        self.conform_to_input_shape(&shape, coerce)
+    fn conform_to_input_shape_py(&self, shape: Shape) -> Result<RearrangeSpec> {
+        self.conform_to_input_shape(&shape)
     }
+
+    #[pyo3(name = "expand_to")]
+    fn expand_to_py(&self, shape: Shape) -> Result<RearrangeSpec> {
+        self.expand_to(&shape)
+    }
+}
+
+pub enum ExpandToSpecOrShape {
+    Spec(RearrangeSpec),
+    SetShape(OpShape),
 }
 
 impl RearrangeSpec {
@@ -900,32 +911,97 @@ impl RearrangeSpec {
             .canonicalize(true)
     }
 
-    pub fn forced_input_shape(
+    pub fn expand_to(&self, shape: &Shape) -> Result<RearrangeSpec> {
+        match self.expand_to_spec_or_shape(shape, true)? {
+            ExpandToSpecOrShape::Spec(spec) => Ok(spec),
+            ExpandToSpecOrShape::SetShape(_) => unreachable!(),
+        }
+    }
+
+    pub fn expand_to_spec_or_shape(
         &self,
-        coerce: bool, /* if true, don't force single int */
-    ) -> OpShape {
-        let mut ret = sv![OpSize::NONE; self.input_ints.len()];
-        for (ints, o) in self.input_ints.iter().zip(&mut ret) {
-            if coerce && ints.len() == 1 {
+        shape: &Shape,
+        error_on_ambiguous: bool,
+    ) -> Result<ExpandToSpecOrShape> {
+        if shape.len() != self.input_ints.len() {
+            bail!(RearrangeSpecError::InputNotConformable {
+                spec: self.clone(),
+                shape: shape.clone(),
+            });
+        }
+
+        let mut out_shape: OpShape = sv![OpSize::NONE; shape.len()];
+        let mut int_sizes = self.int_sizes.clone();
+        for (dim, (inner_ints, &expand_size, out_size)) in
+            izip!(&self.input_ints, shape, &mut out_shape).enumerate()
+        {
+            if inner_ints
+                .iter()
+                .any(|x| self.int_sizes[*x as usize].is_none())
+            {
+                // none size, so this is fine
+                continue;
+            }
+            let current_size = inner_ints
+                .iter()
+                .map(|x| self.int_sizes[*x as usize].unwrap())
+                .product::<usize>();
+            if current_size == expand_size {
+                continue;
+            }
+            if inner_ints.len() == 1 {
+                // change to wild card if single non-matching int
+                int_sizes[inner_ints[0] as usize] = OpSize::NONE;
+                continue;
+            }
+            let inner_int_sizes: Vec<_> = inner_ints
+                .iter()
+                .map(|x| self.int_sizes[*x as usize].unwrap())
+                .collect();
+            let divides_size: Vec<_> = inner_int_sizes
+                .iter()
+                .map(|&x| x != 0 && expand_size % x == 0)
+                .collect();
+            let num_failing_to_divide = divides_size
+                .iter()
+                .map(|divides| (!divides) as usize)
+                .sum::<usize>();
+            if num_failing_to_divide != 1 {
+                if error_on_ambiguous {
+                    bail!(RearrangeSpecError::AmbiguousExpand {
+                        expand_size,
+                        current_size,
+                        inner_ints: inner_ints.clone(),
+                        inner_int_sizes: inner_int_sizes.clone(),
+                        divides_size,
+                        num_failing_to_divide,
+                        dim,
+                        shape: shape.clone(),
+                        spec: self.clone()
+                    })
+                }
+                *out_size = Some(current_size).into();
                 continue;
             }
 
-            *o = ints
+            // if we have exactly 1 symbolic idx, set this size to None
+            let non_dividing_idx = divides_size
                 .iter()
-                .map(|x| -> Option<usize> { self.int_sizes[*x as usize].into() })
-                .collect::<Option<Vec<_>>>()
-                .map(|x| x.iter().product())
-                .into();
+                .enumerate()
+                .find_map(|(i, b)| (!b).then_some(i))
+                .unwrap();
+            int_sizes[inner_ints[non_dividing_idx] as usize] = OpSize::NONE;
+        }
+        if out_shape.iter().any(|x| x.is_some()) {
+            return Ok(ExpandToSpecOrShape::SetShape(out_shape));
         }
 
-        ret
+        Ok(ExpandToSpecOrShape::Spec(
+            Self::new(self.input_ints.clone(), self.output_ints.clone(), int_sizes).unwrap(),
+        ))
     }
 
-    pub fn conform_to_input_shape(
-        &self,
-        shape: &Shape,
-        coerce: bool, /* if false, only change None sizes, if true, change others if unique solution */
-    ) -> Result<RearrangeSpec> {
+    pub fn conform_to_input_shape(&self, shape: &Shape) -> Result<RearrangeSpec> {
         if shape.len() != self.input_ints.len() {
             bail!(RearrangeSpecError::InputNotConformable {
                 spec: self.clone(),
@@ -955,8 +1031,6 @@ impl RearrangeSpec {
             if non_wildcard_size != *l as u64 {
                 if none_indices.len() > 1 {
                     bail!(RearrangeSpecError::TooManyWildcardSizes { spec: self.clone() });
-                } else if coerce && ints.len() == 1 {
-                    int_sizes[ints[0] as usize] = OpSize(*l as u64);
                 } else if *l as u64 % non_wildcard_size != 0 {
                     bail!(RearrangeSpecError::InputNotConformable {
                         spec: self.clone(),
@@ -1127,7 +1201,7 @@ impl RearrangeSpec {
         hasher.finalize().into()
     }
 
-    fn get_unwrapped_sizes(&self) -> Vec<u64> {
+    fn get_unwrapped_sizes(&self) -> Vec<usize> {
         self.int_sizes.iter().map(|x| x.unwrap()).collect()
     }
 
@@ -1154,8 +1228,8 @@ impl RearrangeSpec {
     /// assert_eq!(RearrangeSpec::fuse(&inner, &outer).unwrap(), fused);
     /// ```
     pub fn fuse(inner: &Self, outer: &Self) -> Result<Self> {
-        let mut inner_sizes: Vec<u64> = inner.get_unwrapped_sizes();
-        let mut outer_sizes: Vec<u64> = outer.get_unwrapped_sizes();
+        let mut inner_sizes: Vec<_> = inner.get_unwrapped_sizes();
+        let mut outer_sizes: Vec<_> = outer.get_unwrapped_sizes();
         let mut inner_replaces: HashMap<AxisInt, Vec<AxisInt>> = HashMap::default();
         let mut outer_replaces: HashMap<AxisInt, Vec<AxisInt>> = HashMap::default();
         let mut in_to_out: HashMap<AxisInt, AxisInt> = HashMap::default();
@@ -1368,6 +1442,84 @@ impl RearrangeSpec {
 
         Ok(out)
     }
+
+    /// assumes has size when 'needed' (needed if int ever appears in an inner group with >1 element)
+    /// Can be achieved with conform.
+    pub fn letter_sizes(&self) -> Vec<(String, usize)> {
+        // because to_einops_string uses numerals on all added dims, we only need letter sizes for splits
+
+        self.ints_needing_sizes()
+            .into_iter()
+            .map(|x| {
+                let size_here = self.int_sizes[x as usize].t().unwrap();
+                (ALPHABET[x as usize].to_owned(), size_here)
+            })
+            .collect()
+    }
+
+    /// assumes has size when 'needed' (needed if int ever appears in an inner group with >1 element)
+    /// Can be achieved with conform.
+    pub fn to_einops_string_and_letter_sizes(&self) -> (String, Vec<(String, usize)>) {
+        (self.to_einops_string(), self.letter_sizes())
+    }
+}
+
+#[test]
+fn test_expand_to() {
+    let spec = RearrangeSpec::from_str("a->a").unwrap();
+    assert_eq!(spec.expand_to(&[8].into_iter().collect()).unwrap(), spec);
+    assert_eq!(spec.expand_to(&[2].into_iter().collect()).unwrap(), spec);
+
+    let spec_sized = RearrangeSpec::from_str("a:5->a").unwrap();
+    assert_eq!(
+        spec_sized.expand_to(&[5].into_iter().collect()).unwrap(),
+        spec_sized
+    );
+    assert_eq!(
+        spec_sized.expand_to(&[2].into_iter().collect()).unwrap(),
+        spec
+    );
+    assert_eq!(
+        spec_sized.expand_to(&[6].into_iter().collect()).unwrap(),
+        spec
+    );
+
+    let spec = RearrangeSpec::from_str("(a:5) b (k:2 y:7)-> (k a b) y").unwrap();
+    assert_eq!(
+        spec.expand_to(&[5, 3, 14].into_iter().collect()).unwrap(),
+        spec
+    );
+    assert_eq!(
+        spec.expand_to(&[2, 3, 14].into_iter().collect()).unwrap(),
+        RearrangeSpec::from_str("a b (k:2 y:7)-> (k a b) y").unwrap()
+    );
+    assert_eq!(
+        spec.expand_to(&[7, 3, 14].into_iter().collect()).unwrap(),
+        RearrangeSpec::from_str("a b (k:2 y:7)-> (k a b) y").unwrap()
+    );
+    assert_eq!(
+        spec.expand_to(&[7, 3, 21].into_iter().collect()).unwrap(),
+        RearrangeSpec::from_str("a b (k y:7)-> (k a b) y").unwrap()
+    );
+    assert_eq!(
+        spec.expand_to(&[7, 3, 18].into_iter().collect()).unwrap(),
+        RearrangeSpec::from_str("a b (k:2 y)-> (k a b) y").unwrap()
+    );
+    assert!(spec.expand_to(&[5, 3, 5].into_iter().collect()).is_err());
+    assert!(spec
+        .expand_to(&[2, 5, 3, 14].into_iter().collect())
+        .is_err());
+    assert!(spec.expand_to(&[5, 3, 28].into_iter().collect()).is_err());
+}
+
+fn fancy_size_format<'a, T: IntoIterator<Item = &'a usize>>(x: T) -> String {
+    format!(
+        "[{}]",
+        x.into_iter()
+            .map(|s| SymbolicSizeProduct::from(*s).to_string())
+            .collect::<Vec<_>>()
+            .join(",")
+    )
 }
 
 #[apply(python_error_exception)]
@@ -1396,6 +1548,45 @@ pub enum RearrangeSpecError {
         outer: RearrangeSpec,
     },
 
+    #[error(
+        "expanding to expand_size={fancy_size} (!= current_size={fancy_current_size}) {} {}\n{s}\n{c}\n({e_name})",
+        "was ambiguous for",
+        format!(
+            "inner_ints={:?}, inner_int_sizes={}, divides_size={:?}, num_failing_to_divide={}!=1",
+            inner_ints,
+            fancy_size_format(inner_int_sizes.iter()),
+            divides_size,
+            num_failing_to_divide
+        ),
+        fancy_size = SymbolicSizeProduct::from(*expand_size).to_string(),
+        fancy_current_size =  SymbolicSizeProduct::from(*current_size).to_string(),
+        s = concat!(
+            "ambiguous expansion happens if:\n",
+            "  - there aren't any wildcards\n",
+            "  - there are multiple ints\n",
+            "  - and it isn't the case that exactly one of these ints sizes",
+            " fails to divide the input size ",
+            "(in which case, we can unambiguously replace that size with None)"
+        ),
+        c = format!(
+            "dim={}, shape={:?}, spec={}",
+            dim,
+            fancy_size_format(shape.iter()),
+            spec
+        )
+    )]
+    AmbiguousExpand {
+        expand_size: usize,
+        current_size: usize,
+        inner_ints: RInnerInts,
+        inner_int_sizes: Vec<usize>,
+        divides_size: Vec<bool>,
+        num_failing_to_divide: usize,
+        dim: usize,
+        shape: Shape,
+        spec: RearrangeSpec,
+    },
+
     #[error(
         "Repeated axes={axes:?} unique_count={unique_count} actual_count={actual_count} ({e_name})"
     )]
@@ -1457,7 +1648,7 @@ pub enum PermError {
 
 #[apply(python_error_exception)]
 #[base_error_name(RearrangeParse)]
-#[base_exception(PyValueError)]
+#[base_exception(PyParseError)]
 #[derive(Error, Debug, Clone)]
 pub enum RearrangeParseError {
     // TODO: improve errors
