@@ -33,8 +33,8 @@ pub fn split_prefix(prefix: &str) -> (Vec<String>, String) {
     let n_full_sections = min(
         SECTIONS.len() - 1,
         (0..SECTION_STARTS.len())
-            .map(|i| SECTION_STARTS[i] + SECTIONS[i] <= prefix.len())
-            .count(),
+            .map(|i| (SECTION_STARTS[i] + SECTIONS[i] <= prefix.len()) as usize)
+            .sum(),
     );
 
     (
