@@ -1,7 +1,9 @@
 use circuit_base::{
-    expand_node::ExpandError, generalfunction::GeneralFunctionShapeError,
-    opaque_iterative_matcher::IncorrectOpaqueTypeError, parsing::ParseArgError, ConstructError,
-    TensorEvalError,
+    expand_node::ExpandError,
+    generalfunction::GeneralFunctionShapeError,
+    opaque_iterative_matcher::IncorrectOpaqueTypeError,
+    parsing::{ParseArgError, ParseCircuitError},
+    ConstructError, TensorEvalError,
 };
 use circuit_rewrites::{
     algebraic_rewrite::{DistributeError, PushDownIndexError},
@@ -9,7 +11,7 @@ use circuit_rewrites::{
     deep_rewrite::SimpConfigError,
     scheduled_execution::{SchedulingError, SchedulingOOMError},
 };
-use get_update_node::{iterative_matcher::IterativeMatcherError, sample_transform::SampleError};
+use get_update_node::{iterative_matcher::IterativeMatcherError, sampler::SampleError};
 use nb_operations::{modules::ModuleBindError, nest::NestError};
 use pyo3::{types::PyModule, PyErr, PyResult, Python};
 use rr_util::{
@@ -51,6 +53,7 @@ setup_errors!(
     ExpandError,
     MiscInputError,
     ParseError,
+    ParseCircuitError,
     RearrangeSpecError,
     PermError,
     RearrangeParseError,
