@@ -130,7 +130,7 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_class::<get_update_node::BoundUpdater>()?;
     m.add_class::<get_update_node::Getter>()?;
     m.add_class::<get_update_node::BoundGetter>()?;
-    m.add_class::<get_update_node::Batcher>()?;
+    m.add_class::<get_update_node::Expander>()?;
     m.add_class::<circuit_base::print::PrintOptions>()?;
     m.add_class::<circuit_base::print_html::PrintHtmlOptions>()?;
     m.add_function(wrap_pyfunction!(
@@ -205,6 +205,14 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
 
     m.add_function(wrap_pyfunction!(deep_canonicalize_py, m)?)?;
     m.add_function(wrap_pyfunction!(canonicalize_node_py, m)?)?;
+    m.add_function(wrap_pyfunction!(
+        circuit_rewrites::canonicalize::deep_normalize,
+        m
+    )?)?;
+    m.add_function(wrap_pyfunction!(
+        circuit_rewrites::canonicalize::normalize_node_py,
+        m
+    )?)?;
 
     m.add_function(wrap_pyfunction!(
         circuit_rewrites::batching::batch_to_concat,
@@ -344,7 +352,7 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     )?)?;
     m.add_function(wrap_pyfunction!(rr_util::rrfs::tensor_from_hash, m)?)?;
     m.add_function(wrap_pyfunction!(
-        circuit_base::module::inline_all_modules,
+        circuit_base::module::substitute_all_modules,
         m
     )?)?;
     m.add_function(wrap_pyfunction!(
@@ -360,7 +368,7 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         m
     )?)?;
     m.add_function(wrap_pyfunction!(
-        get_update_node::sample_transform::default_var_matcher,
+        get_update_node::sampler::default_var_matcher,
         m
     )?)?;
     m.add_function(wrap_pyfunction!(
@@ -375,11 +383,11 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         get_update_node::library::apply_in_traversal_py,
         m
     )?)?;
-    m.add_class::<get_update_node::sample_transform::RandomSampleSpec>()?;
-    m.add_class::<get_update_node::sample_transform::RunDiscreteVarAllSpec>()?;
-    m.add_class::<get_update_node::sample_transform::SampleTransform>()?;
+    m.add_class::<get_update_node::sampler::RandomSampleSpec>()?;
+    m.add_class::<get_update_node::sampler::RunDiscreteVarAllSpec>()?;
+    m.add_class::<get_update_node::sampler::Sampler>()?;
     m.add_function(wrap_pyfunction!(
-        get_update_node::sample_transform::factored_cumulant_expectation_rewrite,
+        get_update_node::sampler::factored_cumulant_expectation_rewrite,
         m
     )?)?;
 
